"""2D Animations.

Animations can be used by the :py:class:`~pyglet.sprite.Sprite` class in place
of static images. They are essentially containers for individual image frames,
with a duration per frame. They can be infinitely looping, or stop at the last
frame. You can load Animations from disk, such as from GIF files::

    ani = pyglet.resource.animation('walking.gif')
    sprite = pyglet.sprite.Sprite(img=ani)

Alternatively, you can create your own Animations from a sequence of images
by using the :py:meth:`~Animation.from_image_sequence` method::

    images = [pyglet.resource.image('walk_a.png'),
              pyglet.resource.image('walk_b.png'),
              pyglet.resource.image('walk_c.png')]

    ani = pyglet.image.Animation.from_image_sequence(images, duration=0.1, loop=True)

You can also use an :py:class:`pyglet.image.ImageGrid`, which is iterable::

    sprite_sheet = pyglet.resource.image('my_sprite_sheet.png')
    image_grid = pyglet.image.ImageGrid(sprite_sheet, rows=1, columns=5)

    ani = pyglet.image.Animation.from_image_sequence(image_grid, duration=0.1)

In the above examples, all the Animation Frames have the same duration.
If you wish to adjust this, you can manually create the Animation from a list of
:py:class:`~AnimationFrame`::

    image_a = pyglet.resource.image('walk_a.png')
    image_b = pyglet.resource.image('walk_b.png')
    image_c = pyglet.resource.image('walk_c.png')

    frame_a = pyglet.image.AnimationFrame(image_a, duration=0.1)
    frame_b = pyglet.image.AnimationFrame(image_b, duration=0.2)
    frame_c = pyglet.image.AnimationFrame(image_c, duration=0.1)

    ani = pyglet.image.Animation(frames=[frame_a, frame_b, frame_c])

"""
from __future__ import annotations


from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from typing import Literal
    from pyglet.image import AbstractImage, AbstractImageSequence
    from pyglet.image.atlas import TextureBin


class Animation:
    """Sequence of images with timing information.

    Animations are a collection of :py:class:`~AnimationFrame`s, which are
    simple containers for Image data and duration information.

    If no frames of the animation have a duration of ``None``, the animation
    loops continuously; otherwise the animation stops at the first frame with
    duration of ``None``.
    """

    def __init__(self, frames: list[AnimationFrame]) -> None:
        """Create an animation directly from a list of frames."""
        assert len(frames)
        self.frames = frames

    def add_to_texture_bin(self, texture_bin: TextureBin, border: int = 0) -> None:
        """Add the images of the animation to a :py:class:`~pyglet.image.atlas.TextureBin`.

        The animation frames are modified in-place to refer to the texture bin
        regions. An optional border (in pixels) can be specified to reserve around
        the images that are added to the texture bin.
        """
        for frame in self.frames:
            frame.image = texture_bin.add(frame.image, border)

    def get_transform(self, flip_x: bool = False, flip_y: bool = False,
                      rotate: Literal[0, 90, 180, 270, 360] = 0) -> Animation:
        """Create a copy of this animation, applying a simple transformation.

        The transformation is performed by manipulating the texture coordinates,
        which limits this operation to increments of 90 degrees.

        The transformation is applied around the image's anchor point of
        each frame.  The texture data is shared between the original animation
        and the transformed animation.
        """
        frames = [AnimationFrame(frame.image.get_texture().get_transform(flip_x, flip_y, rotate),
                                 frame.duration) for frame in self.frames]
        return Animation(frames)

    def get_duration(self) -> float:
        """Get the total duration of the animation in seconds."""
        return sum([frame.duration for frame in self.frames if frame.duration is not None])

    def get_max_width(self) -> int:
        """Get the maximum image frame width.

        This method is useful for determining texture space requirements: due
        to the use of ``anchor_x`` the actual required viewing area during
         playback may be larger.
        """
        return max([frame.image.width for frame in self.frames])

    def get_max_height(self) -> int:
        """Get the maximum image frame height.

        This method is useful for determining texture space requirements: due
        to the use of ``anchor_y`` the actual required viewing area during
        playback may be larger.
        """
        return max([frame.image.height for frame in self.frames])

    @classmethod
    def from_image_sequence(cls, sequence: AbstractImageSequence, duration: float, loop: bool = True) -> Animation:
        """Create an animation from a list of images and a per-frame duration."""
        frames = [AnimationFrame(image, duration) for image in sequence]
        if not loop:
            frames[-1].duration = None
        return cls(frames)

    def __repr__(self) -> str:
        return f"Animation(frames={len(self.frames)})"


class AnimationFrame:
    """A single frame of an animation."""

    __slots__ = 'image', 'duration'

    def __init__(self, image: AbstractImage, duration: float | None) -> None:
        """Create an animation frame from an image."""
        self.image = image
        self.duration = duration

    def __repr__(self) -> str:
        return f"AnimationFrame({self.image}, duration={self.duration})"



from __future__ import annotations

import ctypes
import warnings
from ctypes import POINTER, byref, c_buffer, c_char_p, c_int, cast
from typing import TYPE_CHECKING

import pyglet
from pyglet import app
from pyglet.app.xlib import XlibSelectDevice
from pyglet.libs.x11 import xlib
from pyglet.util import asbytes

from . import xlib_vidmoderestore
from .base import Canvas, Display, Screen, ScreenMode

if TYPE_CHECKING:
    from pyglet.gl import Config

try:
    from pyglet.libs.x11 import xinerama

    _have_xinerama = True
except:
    _have_xinerama = False

try:
    from pyglet.libs.x11 import xsync

    _have_xsync = True
except:
    _have_xsync = False

try:
    from pyglet.libs.x11 import xf86vmode

    _have_xf86vmode = True
except:
    _have_xf86vmode = False

try:
    from pyglet.libs.x11 import xrandr

    _have_xrandr = True
except ImportError:
    _have_xrandr = False


class NoSuchDisplayException(Exception):
    pass


# Set up error handler
def _error_handler(display, event):
    # By default, all errors are silently ignored: this has a better chance
    # of working than the default behaviour of quitting ;-)
    #
    # We've actually never seen an error that was our fault; they're always
    # driver bugs (and so the reports are useless).  Nevertheless, set
    # environment variable PYGLET_DEBUG_X11 to 1 to get dumps of the error
    # and a traceback (execution will continue).
    if pyglet.options['debug_x11']:
        event = event.contents
        buf = c_buffer(1024)
        xlib.XGetErrorText(display, event.error_code, buf, len(buf))
        print('X11 error:', buf.value)
        print('   serial:', event.serial)
        print('  request:', event.request_code)
        print('    minor:', event.minor_code)
        print(' resource:', event.resourceid)

        import traceback

        print('Python stack trace (innermost last):')
        traceback.print_stack()
    return 0


_error_handler_ptr = xlib.XErrorHandler(_error_handler)
xlib.XSetErrorHandler(_error_handler_ptr)


class XlibDisplay(XlibSelectDevice, Display):
    _display = None  # POINTER(xlib.Display)

    _x_im = None  # X input method
    # TODO close _x_im when display connection closed.
    _enable_xsync = False
    _screens: list[XlibScreen | XlibScreenXrandr | XlibScreenXinerama]

    def __init__(self, name=None, x_screen=None):
        self._screens = []

        if x_screen is None:
            x_screen = 0

        if isinstance(name, str):
            name = c_char_p(name.encode('ascii'))

        self._display = xlib.XOpenDisplay(name)
        if not self._display:
            raise NoSuchDisplayException(f'Cannot connect to "{name}"')

        screen_count = xlib.XScreenCount(self._display)
        if x_screen >= screen_count:
            raise NoSuchDisplayException(f'Display "{name}" has no screen {x_screen:d}')

        super().__init__()
        self.name = name
        self.x_screen = x_screen

        self._fileno = xlib.XConnectionNumber(self._display)
        self._window_map = {}

        # Initialise XSync
        if _have_xsync:
            event_base = c_int()
            error_base = c_int()
            if xsync.XSyncQueryExtension(self._display, byref(event_base), byref(error_base)):
                major_version = c_int()
                minor_version = c_int()
                if xsync.XSyncInitialize(self._display, byref(major_version), byref(minor_version)):
                    self._enable_xsync = True

        # Add to event loop select list.  Assume we never go away.
        app.platform_event_loop.select_devices.add(self)

    def get_default_screen(self) -> Screen:
        screens = self.get_screens()
        if _have_xrandr:
            for screen in screens:
                if screen.is_primary:
                    return screen

        # Couldn't find a default screen, use the first in the list.
        return self._screens[0]

    def get_screens(self) -> list[XlibScreen]:
        self._screens = []

        # Use XRandr if available, as it appears more maintained and widely supported.
        if _have_xrandr:
            root = xlib.XDefaultRootWindow(self._display)

            res_ptr = xrandr.XRRGetScreenResources(self._display, root)
            if res_ptr:
                primary = xrandr.XRRGetOutputPrimary(self._display, root)

                res = res_ptr.contents
                for i in range(res.noutput):
                    output_id = res.outputs[i]
                    output_info_ptr = xrandr.XRRGetOutputInfo(self._display, res_ptr, output_id)
                    if not output_info_ptr:
                        xrandr.XRRFreeOutputInfo(output_info_ptr)
                        continue

                    output_info: xrandr.XRROutputInfo = output_info_ptr.contents

                    crtc_info_ptr = xrandr.XRRGetCrtcInfo(self._display, res_ptr, output_info.crtc)
                    if not crtc_info_ptr:
                        xrandr.XRRFreeCrtcInfo(crtc_info_ptr)
                        continue

                    crtc_info: xrandr.XRRCrtcInfo = crtc_info_ptr.contents

                    self._screens.append(
                        XlibScreenXrandr(
                            self,
                            crtc_info.x,
                            crtc_info.y,
                            crtc_info.width,
                            crtc_info.height,
                            output_info.crtc,
                            output_id,
                            ctypes.string_at(output_info.name, output_info.nameLen).decode(),
                            output_id == primary,
                        ),
                    )
                    xrandr.XRRFreeCrtcInfo(crtc_info_ptr)
                    xrandr.XRRFreeOutputInfo(output_info_ptr)

            xrandr.XRRFreeScreenResources(res_ptr)

        if not self._screens and _have_xinerama and xinerama.XineramaIsActive(self._display):
            number = c_int()
            infos = xinerama.XineramaQueryScreens(self._display, byref(number))
            infos = cast(infos, POINTER(xinerama.XineramaScreenInfo * number.value)).contents

            using_xinerama = number.value > 1
            for idx, info in enumerate(infos):
                self._screens.append(
                    XlibScreenXinerama(self, info.x_org, info.y_org, info.width, info.height, using_xinerama, idx),
                )
            xlib.XFree(infos)
        elif not self._screens:
            # No xinerama
            screen_info = xlib.XScreenOfDisplay(self._display, self.x_screen)
            screen = XlibScreen(self, 0, 0, screen_info.contents.width, screen_info.contents.height)
            self._screens = [screen]
        return self._screens

    # XlibSelectDevice interface

    def fileno(self) -> int:
        return self._fileno

    def select(self) -> None:
        e = xlib.XEvent()
        while xlib.XPending(self._display):
            xlib.XNextEvent(self._display, e)

            # Key events are filtered by the xlib window event
            # handler so they get a shot at the prefiltered event.
            if e.xany.type not in (xlib.KeyPress, xlib.KeyRelease):
                if xlib.XFilterEvent(e, e.xany.window):
                    continue
            try:
                dispatch = self._window_map[e.xany.window]
            except KeyError:
                continue

            dispatch(e)

    def poll(self) -> int:
        return xlib.XPending(self._display)


class XlibScreen(Screen):
    _initial_mode = None

    def __init__(self, display: XlibDisplay, x: int, y: int, width: int, height: int):
        super().__init__(display, x, y, width, height)

    def get_dpi(self) -> int:
        resource = xlib.XResourceManagerString(self.display._display)
        dpi = 96
        if resource:
            xlib.XrmInitialize()

            db = xlib.XrmGetStringDatabase(resource)
            if db:
                rs_type = c_char_p()
                value = xlib.XrmValue()
                if xlib.XrmGetResource(db, asbytes("Xft.dpi"), asbytes("Xft.dpi"), byref(rs_type), byref(value)):
                    if value.addr and rs_type.value == b'String':
                        dpi = int(value.addr)

                xlib.XrmDestroyDatabase(db)

        return dpi

    def get_scale(self) -> float:
        return self.get_dpi() / 96

    def get_matching_configs(self, template: Config):
        canvas = XlibCanvas(self.display, None)
        configs = template.match(canvas)
        # XXX deprecate
        for config in configs:
            config.screen = self
        return configs

    def get_modes(self) -> list[XlibScreenModeXF86]:
        if not _have_xf86vmode:
            return []

        count = ctypes.c_int()
        info_array = ctypes.POINTER(ctypes.POINTER(xf86vmode.XF86VidModeModeInfo))()
        xf86vmode.XF86VidModeGetAllModeLines(self.display._display, self.display.x_screen, count, info_array)

        depth = xlib.XDefaultDepth(self.display._display, self.display.x_screen)

        # Copy modes out of list and free list
        modes = []
        for i in range(count.value):
            info = xf86vmode.XF86VidModeModeInfo()
            ctypes.memmove(
                ctypes.byref(info),
                ctypes.byref(info_array.contents[i]),
                ctypes.sizeof(info),
            )

            modes.append(XlibScreenModeXF86(self, info, depth))
            if info.privsize:
                xlib.XFree(info.private)
        xlib.XFree(info_array)

        return modes

    def get_mode(self) -> XlibScreenMode:
        modes = self.get_modes()
        if modes:
            return modes[0]
        return None

    def set_mode(self, mode: XlibScreenModeXF86):
        assert mode.screen is self

        if not self._initial_mode:
            self._initial_mode = self.get_mode()
            xlib_vidmoderestore.set_initial_mode(self._initial_mode)

        xf86vmode.XF86VidModeSwitchToMode(self.display._display, self.display.x_screen, mode.info)
        xlib.XFlush(self.display._display)
        xf86vmode.XF86VidModeSetViewPort(self.display._display, self.display.x_screen, 0, 0)
        xlib.XFlush(self.display._display)

        self.width = mode.width
        self.height = mode.height

    def restore_mode(self):
        if self._initial_mode:
            self.set_mode(self._initial_mode)

    def get_display_id(self) -> int:
        # No real unique ID is available, just hash together the properties.
        return hash((self.x, self.y, self.width, self.height))

    def get_monitor_name(self) -> str:
        # No way to get any screen name without XRandr or EDID information.
        return "Unknown"

    def __repr__(self):
        return (
            f"{self.__class__.__name__}(display={self.display!r}, x={self.x}, y={self.y}, "
            f"width={self.width}, height={self.height})"
        )


class XlibScreenXinerama(XlibScreen):
    def __init__(self, display: XlibDisplay, x: int, y: int, width: int, height: int, using_xinerama: bool, idx: int) -> None:
        super().__init__(display, x, y, width, height)
        self._xinerama = using_xinerama
        self.idx = idx

    def get_display_id(self) -> int:
        # No real unique ID is available, just hash together the properties.
        return hash((self.idx, self.x, self.y, self.width, self.height))

    def get_modes(self):
        if self._xinerama:
            # If Xinerama/TwinView is enabled, xf86vidmode's modelines
            # correspond to metamodes, which don't distinguish one screen from
            # another.  XRandR (broken) or NV (complicated) extensions needed.
            return []

        return super().get_modes()


class XlibScreenXrandr(XlibScreen):
    def __init__(
        self,
        display: XlibDisplay,
        x: int,
        y: int,
        width: int,
        height: int,
        crtc_id: int,
        output_id: int,
        name: str,
        is_primary: bool,
    ):
        super().__init__(display, x, y, width, height)
        self.crtc_id = crtc_id
        self.output_id = output_id
        self.name = name
        self._is_primary = is_primary

    @property
    def is_primary(self):
        return self._is_primary

    @staticmethod
    def _get_mode_info(resource: xrandr.XRRScreenResources, rrmode_id: int) -> xrandr.XRRModeInfo | None:
        for i in range(resource.nmode):
            if resource.modes[i].id == rrmode_id:
                return resource.modes[i]

        return None

    def set_mode(self, mode: XlibScreenModeXrandr) -> None:
        assert mode.screen is self

        if not self._initial_mode:
            self._initial_mode = self.get_mode()

        root = xlib.XDefaultRootWindow(self.display._display)
        res_ptr = xrandr.XRRGetScreenResourcesCurrent(self.display._display, root)
        if res_ptr:
            crtc_info_ptr = xrandr.XRRGetCrtcInfo(self.display._display, res_ptr, self.crtc_id)
            if crtc_info_ptr:
                crtc_info = crtc_info_ptr.contents
                status = xrandr.XRRSetCrtcConfig(
                    self.display._display,
                    res_ptr,
                    self.crtc_id,
                    xlib.CurrentTime,
                    crtc_info.x,
                    crtc_info.y,
                    mode.mode_id,
                    crtc_info.rotation,
                    crtc_info.outputs,
                    crtc_info.noutput,
                )
                if status != 0 and pyglet.options['debug_x11']:
                    warnings.warn(f"Could not set screen mode: {status}")
                xlib.XFlush(self.display._display)

                self.width = mode.width
                self.height = mode.height

            xrandr.XRRFreeCrtcInfo(crtc_info_ptr)
        xrandr.XRRFreeScreenResources(res_ptr)

    def get_modes(self) -> list[XlibScreenModeXrandr]:
        modes = []
        root = xlib.XDefaultRootWindow(self.display._display)
        res_ptr = xrandr.XRRGetScreenResourcesCurrent(self.display._display, root)
        output_info_ptr = xrandr.XRRGetOutputInfo(self.display._display, res_ptr, self.output_id)

        res = res_ptr.contents
        output_info = output_info_ptr.contents

        depth = xlib.XDefaultDepth(self.display._display, self.display.x_screen)

        for i in range(output_info_ptr.contents.nmode):
            mode_id = output_info.modes[i]
            xrandr_mode = self._get_mode_info(res, mode_id)
            if xrandr_mode:
                mode = XlibScreenModeXrandr(self, xrandr_mode, mode_id, depth)
                modes.append(mode)

        xrandr.XRRFreeOutputInfo(output_info_ptr)
        xrandr.XRRFreeScreenResources(res_ptr)
        return modes

    def get_mode(self) -> XlibScreenModeXrandr | None:
        # Return the current mode.
        root = xlib.XDefaultRootWindow(self.display._display)
        res_ptr = xrandr.XRRGetScreenResourcesCurrent(self.display._display, root)
        crtc_info_ptr = xrandr.XRRGetCrtcInfo(self.display._display, res_ptr, self.crtc_id)

        crtc_info = crtc_info_ptr.contents
        found_mode = None

        for mode in self.get_modes():
            if crtc_info.mode == mode.mode_id:
                found_mode = mode
                break

        xrandr.XRRFreeCrtcInfo(crtc_info_ptr)
        xrandr.XRRFreeScreenResources(res_ptr)
        return found_mode

    def get_display_id(self) -> int:
        return self.output_id

    def get_monitor_name(self) -> str:
        return self.name


class XlibScreenMode(ScreenMode):
    def __init__(self, screen: XlibScreen, width: int, height: int, rate: int, depth: int):
        super().__init__(screen)
        self.width = width
        self.height = height
        self.rate = rate
        self.depth = depth


class XlibScreenModeXF86(XlibScreenMode):
    def __init__(self, screen: XlibScreen, info: xf86vmode.XF86VidModeModeInfo, depth: int) -> None:
        self.info = info
        width = info.hdisplay
        height = info.vdisplay
        rate = round((info.dotclock * 1000) / (info.htotal * info.vtotal))
        super().__init__(screen, width, height, rate, depth)

    def __repr__(self) -> str:
        return f'XlibScreenMode(width={self.width!r}, height={self.height!r}, depth={self.depth!r}, rate={self.rate})'


class XlibScreenModeXrandr(XlibScreenMode):
    def __init__(self, screen: XlibScreen, mode_info: xrandr.XRRModeInfo, mode_id: int, depth: int) -> None:
        self.mode_id = mode_id
        super().__init__(screen, mode_info.width, mode_info.height, self._calculate_refresh_rate(mode_info), depth)

    @staticmethod
    def _calculate_refresh_rate(mode_info: xrandr.XRRModeInfo) -> int:
        if mode_info.hTotal > 0 and mode_info.vTotal > 0:
            return round(mode_info.dotClock / (mode_info.hTotal * mode_info.vTotal))
        return 0

    def __repr__(self) -> str:
        return f'XlibScreenMode(width={self.width!r}, height={self.height!r}, depth={self.depth!r}, rate={self.rate})'


class XlibCanvas(Canvas):  # noqa: D101
    display: XlibDisplay

    def __init__(self, display: XlibDisplay, x_window: xlib.Window) -> None:  # noqa: D107
        super().__init__(display)
        self.x_window = x_window



"""Display and screen management.

Rendering is performed on the content area of a :class:`pyglet.window.Window`,
or an entire screen.

Windows must belong to a :class:`Display`. On Microsoft Windows and macOS,
there is only one display, which can be obtained with :func:`get_display`.
Linux supports multiple displays, corresponding to discrete X11 display
connections and screens.  :func:`get_display` on Linux returns the default
display and screen 0 (``localhost:0.0``); if a particular screen or display is
required then :class:`Display` can be instantiated directly.

Within a display one or more screens are attached.  A :class:`Screen` often
corresponds to a physical attached monitor, however a monitor or projector set
up to clone another screen will not be listed.  Use :meth:`Display.get_screens`
to get a list of the attached screens; these can then be queried for their
sizes and virtual positions on the desktop.

The size of a screen is determined by its current mode, which can be changed
by the application; see the documentation for :class:`Screen`.

.. versionadded:: 1.2
"""  # noqa: I002

import sys
import weakref

_is_pyglet_doc_run = hasattr(sys, "is_pyglet_doc_run") and sys.is_pyglet_doc_run


if _is_pyglet_doc_run:
    from pyglet.display.base import Display, Screen, Canvas, ScreenMode
else:
    from pyglet import compat_platform, options
    if options['headless']:
        from pyglet.display.headless import HeadlessDisplay as Display
        from pyglet.display.headless import HeadlessScreen as Screen
        from pyglet.display.headless import HeadlessCanvas as Canvas
    elif compat_platform == 'darwin':
        from pyglet.display.cocoa import CocoaDisplay as Display
        from pyglet.display.cocoa import CocoaScreen as Screen
        from pyglet.display.cocoa import CocoaCanvas as Canvas
    elif compat_platform in ('win32', 'cygwin'):
        from pyglet.display.win32 import Win32Display as Display
        from pyglet.display.win32 import Win32Screen as Screen
        from pyglet.display.win32 import Win32Canvas as Canvas
    elif compat_platform == 'linux':
        from pyglet.display.xlib import XlibDisplay as Display
        from pyglet.display.xlib import XlibScreen as Screen
        from pyglet.display.xlib import XlibCanvas as Canvas
    else:
        msg = f"A display interface for '{compat_platform}' is not yet implemented."
        raise NotImplementedError(msg)


_displays: weakref.WeakSet = weakref.WeakSet()
"""Set of all open displays.  Instances of :class:`Display` are automatically
added to this set upon construction.  The set uses weak references, so displays
are removed from the set when they are no longer referenced.
"""


def get_display() -> Display:
    """Get the default display device.

    If there is already a :class:`Display` connection, that display will be
    returned. Otherwise, a default :class:`Display` is created and returned.
    If multiple display connections are active, an arbitrary one is returned.

    .. versionadded:: 1.2
    """
    # If there are existing displays, return one of them arbitrarily.
    for display in _displays:
        return display

    # Otherwise, create a new display and return it.
    return Display()


__all__ = ['Display', 'Screen', 'Canvas', 'ScreenMode', 'get_display']



from ctypes import *
from pyglet.libs.egl import egl
from pyglet.libs.egl.lib import link_EGL as _link_function


EGL_PLATFORM_GBM_MESA = 12759
EGL_PLATFORM_DEVICE_EXT = 12607
EGLDeviceEXT = POINTER(None)

eglGetPlatformDisplayEXT = _link_function('eglGetPlatformDisplayEXT', egl.EGLDisplay, [egl.EGLenum, POINTER(None), POINTER(egl.EGLint)], None)
eglCreatePlatformWindowSurfaceEXT = _link_function('eglCreatePlatformWindowSurfaceEXT', egl.EGLSurface, [egl.EGLDisplay, egl.EGLConfig, POINTER(None), POINTER(egl.EGLAttrib)], None)
eglQueryDevicesEXT = _link_function('eglQueryDevicesEXT', egl.EGLBoolean, [egl.EGLint, POINTER(EGLDeviceEXT), POINTER(egl.EGLint)], None)


__all__ = ['EGL_PLATFORM_DEVICE_EXT', 'EGL_PLATFORM_GBM_MESA',
           'EGLDeviceEXT', 'eglGetPlatformDisplayEXT', 'eglCreatePlatformWindowSurfaceEXT',
           'eglQueryDevicesEXT']



import re
import struct

from concurrent.futures import ProcessPoolExecutor

from pyglet.util import DecodeException
from pyglet.media.codecs.base import StreamingSource, AudioData, AudioFormat, StaticSource
from pyglet.media.codecs import MediaDecoder


# This module uses code from Project Nayuki:
# https://www.nayuki.io/page/simple-flac-implementation


class FLACDecodeException(DecodeException):
    pass


FIXED_PREDICTION_COEFFICIENTS = ((), (1,), (2, -1), (3, -3, 1), (4, -6, 4, -1))


def _initializer(filename, numchannels, sample_size):
    """Function to open a copy of the file locally in each Process."""
    global _file
    global _numchannels
    global _sample_size
    _file = open(filename, 'rb')
    _numchannels = numchannels
    _sample_size = sample_size


def decode_frame(offset):
    try:
        # Read a ton of header fields, and ignore most of them
        numchannels = _numchannels
        sample_size = _sample_size
        _file.seek(offset)
        inp = BitInputStream(_file)

        temp = inp.read_byte()
        if temp == -1:
            return False
        sync = temp << 6 | inp.read_uint(6)
        if sync != 0x3FFE:
            raise FLACDecodeException("Sync code expected")

        inp.read_uint(1)
        inp.read_uint(1)
        blocksizecode = inp.read_uint(4)
        sampleratecode = inp.read_uint(4)
        chanasgn = inp.read_uint(4)
        inp.read_uint(3)
        inp.read_uint(1)

        temp = inp.read_uint(8)
        while temp >= 0b11000000:
            inp.read_uint(8)
            temp = (temp << 1) & 0xFF

        if blocksizecode == 1:
            blocksize = 192
        elif 2 <= blocksizecode <= 5:
            blocksize = 576 << blocksizecode - 2
        elif blocksizecode == 6:
            blocksize = inp.read_uint(8) + 1
        elif blocksizecode == 7:
            blocksize = inp.read_uint(16) + 1
        elif 8 <= blocksizecode <= 15:
            blocksize = 256 << (blocksizecode - 8)
        else:
            raise FLACDecodeException("Invalid Blocksize")

        if sampleratecode == 12:
            inp.read_uint(8)
        elif sampleratecode in (13, 14):
            inp.read_uint(16)

        crc8 = inp.read_byte()

        # Decode each channel's subframe, then skip footer
        if 0 <= chanasgn <= 7:
            samples = [decode_subframe(inp, blocksize, sample_size) for _ in range(chanasgn + 1)]
        elif 8 <= chanasgn <= 10:
            temp0 = decode_subframe(inp, blocksize, sample_size + (1 if (chanasgn == 9) else 0))
            temp1 = decode_subframe(inp, blocksize, sample_size + (0 if (chanasgn == 9) else 1))
            if chanasgn == 8:
                for i in range(blocksize):
                    temp1[i] = temp0[i] - temp1[i]
            elif chanasgn == 9:
                for i in range(blocksize):
                    temp0[i] += temp1[i]
            elif chanasgn == 10:
                for i in range(blocksize):
                    side = temp1[i]
                    right = temp0[i] - (side >> 1)
                    temp1[i] = right
                    temp0[i] = right + side
            samples = [temp0, temp1]
        else:
            raise FLACDecodeException("Reserved channel assignment")

        inp.align_to_byte()
        inp.read_byte()
        inp.read_byte()

        # Write the decoded samples
        fmt = str(blocksize * numchannels) + 'h'
        if sample_size == 8:
            return struct.pack(fmt, *(samples[i][j] + 128 for j in range(blocksize) for i in range(numchannels)))
        else:
            return struct.pack(fmt, *(samples[i][j] for j in range(blocksize) for i in range(numchannels)))
    except:
        return b""


def decode_subframe(inp, blocksize, sampledepth):
    inp.read_uint(1)
    subframe_type = inp.read_uint(6)
    shift = inp.read_uint(1)
    if shift == 1:
        while inp.read_uint(1) == 0:
            shift += 1
    sampledepth -= shift

    if subframe_type == 0:  # Constant coding
        result = [inp.read_signed_int(sampledepth)] * blocksize
    elif subframe_type == 1:  # Verbatim coding
        result = [inp.read_signed_int(sampledepth) for _ in range(blocksize)]
    elif 8 <= subframe_type <= 12:
        result = decode_fixed_prediction_subframe(inp, subframe_type - 8, blocksize, sampledepth)
    elif 32 <= subframe_type <= 63:
        result = decode_linear_predictive_coding_subframe(inp, subframe_type - 31, blocksize, sampledepth)
    else:
        raise FLACDecodeException("Reserved subframe type")
    return [(v << shift) for v in result]


def decode_fixed_prediction_subframe(inp, predorder, blocksize, sampledepth):
    result = [inp.read_signed_int(sampledepth) for _ in range(predorder)]
    decode_residuals(inp, blocksize, result)
    restore_linear_prediction(result, FIXED_PREDICTION_COEFFICIENTS[predorder], 0)
    return result


def decode_linear_predictive_coding_subframe(inp, lpcorder, blocksize, sampledepth):
    result = [inp.read_signed_int(sampledepth) for _ in range(lpcorder)]
    precision = inp.read_uint(4) + 1
    shift = inp.read_signed_int(5)
    coefs = [inp.read_signed_int(precision) for _ in range(lpcorder)]
    decode_residuals(inp, blocksize, result)
    restore_linear_prediction(result, coefs, shift)
    return result


def decode_residuals(inp, blocksize, result):
    method = inp.read_uint(2)
    if method >= 2:
        raise FLACDecodeException("Reserved residual coding method")
    parambits = [4, 5][method]
    escapeparam = [0xF, 0x1F][method]

    partitionorder = inp.read_uint(4)
    numpartitions = 1 << partitionorder
    if blocksize % numpartitions != 0:
        raise FLACDecodeException("Block size not divisible by number of Rice partitions")

    for i in range(numpartitions):
        count = blocksize >> partitionorder
        if i == 0:
            count -= len(result)
        param = inp.read_uint(parambits)
        if param < escapeparam:
            result.extend(inp.read_rice_signed_int(param) for _ in range(count))
        else:
            numbits = inp.read_uint(5)
            result.extend(inp.read_signed_int(numbits) for _ in range(count))


def restore_linear_prediction(result, coefs, shift):
    for i in range(len(coefs), len(result)):
        result[i] += sum((result[i - 1 - j] * c) for (j, c) in enumerate(coefs)) >> shift


class BitInputStream:
    __slots__ = 'inp', 'bitbuffer', 'bitbufferlen'

    def __init__(self, inp):
        self.inp = inp
        self.bitbuffer = 0
        self.bitbufferlen = 0

    def align_to_byte(self):
        self.bitbufferlen -= self.bitbufferlen % 8

    def read_byte(self):
        if self.bitbufferlen >= 8:
            return self.read_uint(8)
        else:
            result = self.inp.read(1)
            if len(result) == 0:
                return -1
            return result[0]

    def read_uint(self, n):
        bitbuffer = self.bitbuffer
        bitbufferlen = self.bitbufferlen

        while bitbufferlen < n:
            temp = self.inp.read(1)
            if len(temp) == 0:
                raise EOFError()
            temp = temp[0]
            bitbuffer = (bitbuffer << 8) | temp
            bitbufferlen += 8
        bitbufferlen -= n
        result = (bitbuffer >> bitbufferlen) & ((1 << n) - 1)
        bitbuffer &= (1 << bitbufferlen) - 1

        self.bitbuffer = bitbuffer
        self.bitbufferlen = bitbufferlen
        return result

    def read_signed_int(self, n):
        temp = self.read_uint(n)
        temp -= (temp >> (n - 1)) << n
        return temp

    def read_rice_signed_int(self, param):
        read_uint = self.read_uint
        val = 0
        while read_uint(1) == 0:
            val += 1
        val = (val << param) | read_uint(param)
        return (val >> 1) ^ -(val & 1)

    def close(self):
        self.inp.close()


def _find_frame_indicies(data):
    magic1 = struct.pack('BB', int('11111111', 2), int('11111000', 2))
    magic2 = struct.pack('BB', int('11111111', 2), int('11111001', 2))
    indices1 = [m.start() for m in re.finditer(magic1, data)]
    indices2 = [m.start() for m in re.finditer(magic2, data)]
    indices = indices1 + indices2
    indices.sort()
    return indices


#########################################
#   Source class:
#########################################
class FLACSource(StreamingSource):
    def __init__(self, filename, file=None):
        if file is None:
            file = open(filename, 'rb')

        self.frame_indices = _find_frame_indicies(file.read())
        self.frame_index = 0
        file.seek(0)
        self._file = BitInputStream(file)
        self._filename = filename

        if file.read(4) != b'fLaC':
            raise FLACDecodeException("Does not appear to be a FLAC file.")

        samplerate = numchannels = bits_per_sample = numsamples = None

        last = False
        while not last:
            last = self._file.read_uint(1) != 0
            block_type = self._file.read_uint(7)
            length = self._file.read_uint(24)
            if block_type == 0:  # Stream info block
                min_block_size = self._file.read_uint(16)  # in samples
                max_block_size = self._file.read_uint(16)  # in samples
                min_frame_size = self._file.read_uint(24)  # in bytes,  0 == unknown
                max_frame_size = self._file.read_uint(24)  # in bytes,  0 == unknown
                samplerate = self._file.read_uint(20)
                numchannels = self._file.read_uint(3) + 1
                bits_per_sample = self._file.read_uint(5) + 1
                numsamples = self._file.read_uint(36)
                self._file.read_uint(128)
            else:   # Skip other metadata blocks
                for i in range(length):
                    self._file.read_byte()

        for offset in self.frame_indices:
            if offset < self._file.inp.tell():
                self.frame_indices.remove(offset)

        if not samplerate:
            raise FLACDecodeException("Stream info metadata block absent")
        if bits_per_sample % 8 != 0:
            raise FLACDecodeException("Sample depth {} not supported".format(bits_per_sample))

        self._numsamples = numsamples
        self._duration = float(numsamples) / samplerate / 60
        self.audio_format = AudioFormat(numchannels, bits_per_sample, samplerate)

        self.executor = ProcessPoolExecutor(max_workers=3, initializer=_initializer,
                                            initargs=(filename, numchannels, bits_per_sample))

    def __del__(self):
        self._file.close()

    def get_audio_data(self, num_bytes, compensation_time=0.0):

        data = b''

        while not data and (self.frame_index <= len(self.frame_indices)):
            offset = self.frame_indices[self.frame_index]
            future = self.executor.submit(decode_frame, offset)
            data += future.result()
            self.frame_index += 1

        if not data:
            return None

        return AudioData(data, len(data), timestamp=0.1, duration=0.5, events=[])

    def save(self, filename="output.wav"):
        with open(filename, 'wb') as out:
            numsamples = self._numsamples
            numchannels = self.audio_format.channels
            samplerate = self.audio_format.sample_rate
            sample_size = self.audio_format.sample_size

            # Start writing WAV file headers
            sampledatalen = numsamples * numchannels * (sample_size // 8)

            out.write(b"RIFF")
            out.write(struct.pack("<I", sampledatalen + 36))
            out.write(b"WAVE")
            out.write(b"fmt ")
            out.write(struct.pack("<IHHIIHH",
                                  16,
                                  0x0001,
                                  numchannels,
                                  samplerate,
                                  samplerate * numchannels * (sample_size // 8),
                                  numchannels * (sample_size // 8),
                                  sample_size))
            out.write(b"data")
            out.write(struct.pack("<I", sampledatalen))

            import time
            start = time.time()
            futures = [self.executor.submit(decode_frame, offset) for offset in self.frame_indices]
            print("elapsed future making", time.time() - start)

            while futures:
                data = futures.pop(0).result()
                out.write(data)

    def seek(self, timestamp):
        pass


#########################################
#   Decoder class:
#########################################
class FLACDecoder(MediaDecoder):

    def get_file_extensions(self):
        return '.flac',

    def decode(self, file, filename, streaming=True):
        if streaming:
            return FLACSource(filename, file)
        else:
            return StaticSource(FLACSource(filename, file))


def get_decoders():
    return [FLACDecoder()]


def get_encoders():
    return []



from pyglet.media.exceptions import MediaException


class DirectSoundException(MediaException):
    pass


class DirectSoundNativeError(DirectSoundException):
    def __init__(self, hresult):
        self.hresult = hresult

    def __repr__(self):
        return "{}: Error {}".format(self.__class__.__name__, self.hresult)



# Note: The display mode API used here is Mac OS 10.6 only.
from __future__ import annotations

from ctypes import c_uint32, c_void_p, byref

from .base import Canvas, Display, Screen, ScreenMode
from pyglet.libs.darwin.cocoapy import CGDirectDisplayID, quartz, cf, ObjCClass, get_NSString
from pyglet.libs.darwin.cocoapy import cfstring_to_string, cfarray_to_list
from pyglet.libs.darwin import NSDeviceResolution


NSScreen = ObjCClass('NSScreen')


class CocoaDisplay(Display):

    def get_screens(self):
        maxDisplays = 256
        activeDisplays = (CGDirectDisplayID * maxDisplays)()
        count = c_uint32()
        quartz.CGGetActiveDisplayList(maxDisplays, activeDisplays, byref(count))
        return [CocoaScreen(self, displayID) for displayID in list(activeDisplays)[: count.value]]

    def get_default_screen(self) -> Screen:
        main_id = quartz.CGMainDisplayID()

        screens = self.get_screens()
        for screen in screens:
            if screen.get_display_id() == main_id:
                return screen

        return screens[0]

class CocoaScreen(Screen):

    def __init__(self, display, displayID):
        bounds = quartz.CGDisplayBounds(displayID)
        # FIX ME:
        # Probably need to convert the origin coordinates depending on context:
        # http://www.cocoabuilder.com/archive/cocoa/233492-ns-cg-rect-conversion-and-screen-coordinates.html
        x, y = bounds.origin.x, bounds.origin.y
        width, height = bounds.size.width, bounds.size.height
        super().__init__(display, int(x), int(y), int(width), int(height))
        self._cg_display_id = displayID
        # Save the default mode so we can restore to it.
        self._default_mode = self.get_mode()
        self._ns_screen = self.get_nsscreen()
        self._friendly_name = "Unknown"
        if self._ns_screen is not None:
            screen_name = self._ns_screen.localizedName()
            if screen_name:
                self._friendly_name = cfstring_to_string(screen_name)

    def get_nsscreen(self):
        """Returns the NSScreen instance that matches our CGDirectDisplayID."""
        # Get a list of all currently active NSScreens and then search through
        # them until we find one that matches our CGDisplayID.
        screen_array = NSScreen.screens()
        count = screen_array.count()
        for i in range(count):
            nsscreen = screen_array.objectAtIndex_(i)
            screenInfo = nsscreen.deviceDescription()
            displayID = screenInfo.objectForKey_(get_NSString('NSScreenNumber'))
            displayID = displayID.intValue()
            if displayID == self._cg_display_id:
                return nsscreen
        return None

    def get_dpi(self):
        desc = self._ns_screen.deviceDescription()
        rsize = desc.objectForKey_(NSDeviceResolution).sizeValue()
        return int(rsize.width)

    def get_scale(self):
        ratio = 1.0
        if self._ns_screen:
            pts = self._ns_screen.frame()
            pixels = self._ns_screen.convertRectToBacking_(pts)
            ratio = pixels.size.width / pts.size.width
        else:
            print("Could not initialize NSScreen to retrieve DPI. Using default.")

        return ratio

    def get_matching_configs(self, template):
        canvas = CocoaCanvas(self.display, self, None)
        return template.match(canvas)

    def get_modes(self):
        cgmodes = c_void_p(quartz.CGDisplayCopyAllDisplayModes(self._cg_display_id, None))
        modes = [CocoaScreenMode(self, cgmode) for cgmode in cfarray_to_list(cgmodes)]
        cf.CFRelease(cgmodes)
        return modes

    def get_mode(self):
        cgmode = c_void_p(quartz.CGDisplayCopyDisplayMode(self._cg_display_id))
        mode = CocoaScreenMode(self, cgmode)
        quartz.CGDisplayModeRelease(cgmode)
        return mode

    def set_mode(self, mode):
        assert mode.screen is self
        quartz.CGDisplayCapture(self._cg_display_id)
        quartz.CGDisplaySetDisplayMode(self._cg_display_id, mode.cgmode, None)
        self.width = mode.width
        self.height = mode.height

    def restore_mode(self):
        match_attrs = ['width', 'height', 'depth', 'rate']
        current_mode = self.get_mode()
        if all(getattr(current_mode, attr) == getattr(self._default_mode, attr) for
               attr in match_attrs):
            # Already in default mode
            return
        quartz.CGDisplaySetDisplayMode(self._cg_display_id, self._default_mode.cgmode, None)
        quartz.CGDisplayRelease(self._cg_display_id)

    def capture_display(self):
        quartz.CGDisplayCapture(self._cg_display_id)

    def release_display(self):
        quartz.CGDisplayRelease(self._cg_display_id)

    def get_display_id(self) -> str:
        """Get a unique identifier for the screen."""
        return self._cg_display_id

    def get_monitor_name(self) -> str:
        """Get a friendly name, if available.

        For example, the make and model of the screen: Dell S2716DG
        """
        return self._friendly_name

class CocoaScreenMode(ScreenMode):

    def __init__(self, screen, cgmode):
        super().__init__(screen)
        quartz.CGDisplayModeRetain(cgmode)
        self.cgmode = cgmode
        self.width = int(quartz.CGDisplayModeGetWidth(cgmode))
        self.height = int(quartz.CGDisplayModeGetHeight(cgmode))
        self.depth = self.getBitsPerPixel(cgmode)
        self.rate = quartz.CGDisplayModeGetRefreshRate(cgmode)

    def __del__(self):
        quartz.CGDisplayModeRelease(self.cgmode)
        self.cgmode = None

    def getBitsPerPixel(self, cgmode):
        # from /System/Library/Frameworks/IOKit.framework/Headers/graphics/IOGraphicsTypes.h
        IO8BitIndexedPixels = "PPPPPPPP"
        IO16BitDirectPixels = "-RRRRRGGGGGBBBBB"
        IO32BitDirectPixels = "--------RRRRRRRRGGGGGGGGBBBBBBBB"

        cfstring = c_void_p(quartz.CGDisplayModeCopyPixelEncoding(cgmode))
        pixelEncoding = cfstring_to_string(cfstring)
        cf.CFRelease(cfstring)

        if pixelEncoding == IO8BitIndexedPixels: return 8
        if pixelEncoding == IO16BitDirectPixels: return 16
        if pixelEncoding == IO32BitDirectPixels: return 32
        return 0


class CocoaCanvas(Canvas):

    def __init__(self, display, screen, nsview):
        super().__init__(display)
        self.screen = screen
        self.nsview = nsview



from ctypes import *

import pyglet.lib


_lib = pyglet.lib.load_library('gbm')


class c_void(Structure):
    # c_void_p is a buggy return type, converting to int, so
    # POINTER(None) == c_void_p is actually written as
    # POINTER(c_void), so it can be treated as a real pointer.
    _fields_ = [('dummy', c_int)]


class struct_gbm_bo(Structure):
    _fields_ = [('_opaque_struct', c_int)]


class struct_gbm_bo_handle(Union):
    _fields_ = [('_opaque_struct', c_int)]


class struct_gbm_device(Structure):
    _fields_ = [('_opaque_struct', c_int)]


class struct_gbm_surface(Structure):
    _fields_ = [('_opaque_struct', c_int)]


class struct_gbm_format_name_desc(Structure):
    _fields_ = [('_opaque_struct', c_int)]


__GBM__ = 1  # /usr/include/gbm.h:31

# enum
# GBM_BO_FORMAT_XRGB8888
# GBM_BO_FORMAT_ARGB8888

GBM_BO_USE_SCANOUT = (1 << 0)
GBM_BO_USE_CURSOR = (1 << 1)
GBM_BO_USE_CURSOR_64X64 = GBM_BO_USE_CURSOR
GBM_BO_USE_RENDERING = (1 << 2)

GBM_BO_USE_WRITE = (1 << 3)
GBM_BO_USE_LINEAR = (1 << 4)
GBM_BO_USE_PROTECTED = (1 << 5)

GBM_BO_TRANSFER_READ = (1 << 0)
GBM_BO_TRANSFER_WRITE = (1 << 1)
GBM_BO_TRANSFER_READ_WRITE = (GBM_BO_TRANSFER_READ | GBM_BO_TRANSFER_WRITE)


# TODO: pre-calculate all of the fourcc values
def __gbm_fourcc_code(a, b, c, d):
    a, b, c, d = ord(a), ord(b), ord(c), ord(d)
    return c_uint32(a).value | (c_uint32(b).value << 8) | (c_uint32(c).value << 16) | (c_uint32(d).value << 24)


GBM_FORMAT_BIG_ENDIAN = (1 << 31)
GBM_FORMAT_C8 = __gbm_fourcc_code('C', '8', ' ', ' ')
GBM_FORMAT_R8 = __gbm_fourcc_code('R', '8', ' ', ' ')
GBM_FORMAT_R16 = __gbm_fourcc_code('R', '1', '6', ' ')

GBM_FORMAT_GR88 = __gbm_fourcc_code('G', 'R', '8', '8')

GBM_FORMAT_RG1616 = __gbm_fourcc_code('R', 'G', '3', '2')
GBM_FORMAT_GR1616 = __gbm_fourcc_code('G', 'R', '3', '2')

GBM_FORMAT_RGB332 = __gbm_fourcc_code('R', 'G', 'B', '8')
GBM_FORMAT_BGR233 = __gbm_fourcc_code('B', 'G', 'R', '8')

GBM_FORMAT_XRGB4444 = __gbm_fourcc_code('X', 'R', '1', '2')
GBM_FORMAT_XBGR4444 = __gbm_fourcc_code('X', 'B', '1', '2')
GBM_FORMAT_RGBX4444 = __gbm_fourcc_code('R', 'X', '1', '2')
GBM_FORMAT_BGRX4444 = __gbm_fourcc_code('B', 'X', '1', '2')

GBM_FORMAT_ARGB4444 = __gbm_fourcc_code('A', 'R', '1', '2')
GBM_FORMAT_ABGR4444 = __gbm_fourcc_code('A', 'B', '1', '2')
GBM_FORMAT_RGBA4444 = __gbm_fourcc_code('R', 'A', '1', '2')
GBM_FORMAT_BGRA4444 = __gbm_fourcc_code('B', 'A', '1', '2')

GBM_FORMAT_XRGB1555 = __gbm_fourcc_code('X', 'R', '1', '5')
GBM_FORMAT_XBGR1555 = __gbm_fourcc_code('X', 'B', '1', '5')
GBM_FORMAT_RGBX5551 = __gbm_fourcc_code('R', 'X', '1', '5')
GBM_FORMAT_BGRX5551 = __gbm_fourcc_code('B', 'X', '1', '5')

GBM_FORMAT_ARGB1555 = __gbm_fourcc_code('A', 'R', '1', '5')
GBM_FORMAT_ABGR1555 = __gbm_fourcc_code('A', 'B', '1', '5')
GBM_FORMAT_RGBA5551 = __gbm_fourcc_code('R', 'A', '1', '5')
GBM_FORMAT_BGRA5551 = __gbm_fourcc_code('B', 'A', '1', '5')

GBM_FORMAT_RGB565 = __gbm_fourcc_code('R', 'G', '1', '6')
GBM_FORMAT_BGR565 = __gbm_fourcc_code('B', 'G', '1', '6')

GBM_FORMAT_RGB888 = __gbm_fourcc_code('R', 'G', '2', '4')
GBM_FORMAT_BGR888 = __gbm_fourcc_code('B', 'G', '2', '4')

GBM_FORMAT_XRGB8888 = __gbm_fourcc_code('X', 'R', '2', '4')
GBM_FORMAT_XBGR8888 = __gbm_fourcc_code('X', 'B', '2', '4')
GBM_FORMAT_RGBX8888 = __gbm_fourcc_code('R', 'X', '2', '4')
GBM_FORMAT_BGRX8888 = __gbm_fourcc_code('B', 'X', '2', '4')

GBM_FORMAT_ARGB8888 = __gbm_fourcc_code('A', 'R', '2', '4')
GBM_FORMAT_ABGR8888 = __gbm_fourcc_code('A', 'B', '2', '4')
GBM_FORMAT_RGBA8888 = __gbm_fourcc_code('R', 'A', '2', '4')
GBM_FORMAT_BGRA8888 = __gbm_fourcc_code('B', 'A', '2', '4')

GBM_FORMAT_XRGB2101010 = __gbm_fourcc_code('X', 'R', '3', '0')
GBM_FORMAT_XBGR2101010 = __gbm_fourcc_code('X', 'B', '3', '0')
GBM_FORMAT_RGBX1010102 = __gbm_fourcc_code('R', 'X', '3', '0')
GBM_FORMAT_BGRX1010102 = __gbm_fourcc_code('B', 'X', '3', '0')

GBM_FORMAT_ARGB2101010 = __gbm_fourcc_code('A', 'R', '3', '0')
GBM_FORMAT_ABGR2101010 = __gbm_fourcc_code('A', 'B', '3', '0')
GBM_FORMAT_RGBA1010102 = __gbm_fourcc_code('R', 'A', '3', '0')
GBM_FORMAT_BGRA1010102 = __gbm_fourcc_code('B', 'A', '3', '0')

GBM_FORMAT_XBGR16161616F = __gbm_fourcc_code('X', 'B', '4', 'H')
GBM_FORMAT_ABGR16161616F = __gbm_fourcc_code('A', 'B', '4', 'H')

GBM_FORMAT_YUYV = __gbm_fourcc_code('Y', 'U', 'Y', 'V')
GBM_FORMAT_YVYU = __gbm_fourcc_code('Y', 'V', 'Y', 'U')
GBM_FORMAT_UYVY = __gbm_fourcc_code('U', 'Y', 'V', 'Y')
GBM_FORMAT_VYUY = __gbm_fourcc_code('V', 'Y', 'U', 'Y')

GBM_FORMAT_AYUV = __gbm_fourcc_code('A', 'Y', 'U', 'V')
GBM_FORMAT_NV12 = __gbm_fourcc_code('N', 'V', '1', '2')
GBM_FORMAT_NV21 = __gbm_fourcc_code('N', 'V', '2', '1')
GBM_FORMAT_NV16 = __gbm_fourcc_code('N', 'V', '1', '6')
GBM_FORMAT_NV61 = __gbm_fourcc_code('N', 'V', '6', '1')

GBM_FORMAT_YUV410 = __gbm_fourcc_code('Y', 'U', 'V', '9')
GBM_FORMAT_YVU410 = __gbm_fourcc_code('Y', 'V', 'U', '9')
GBM_FORMAT_YUV411 = __gbm_fourcc_code('Y', 'U', '1', '1')
GBM_FORMAT_YVU411 = __gbm_fourcc_code('Y', 'V', '1', '1')
GBM_FORMAT_YUV420 = __gbm_fourcc_code('Y', 'U', '1', '2')
GBM_FORMAT_YVU420 = __gbm_fourcc_code('Y', 'V', '1', '2')
GBM_FORMAT_YUV422 = __gbm_fourcc_code('Y', 'U', '1', '6')
GBM_FORMAT_YVU422 = __gbm_fourcc_code('Y', 'V', '1', '6')
GBM_FORMAT_YUV444 = __gbm_fourcc_code('Y', 'U', '2', '4')
GBM_FORMAT_YVU444 = __gbm_fourcc_code('Y', 'V', '2', '4')


# /usr/include/gbm.h:257
gbm_device_get_fd = _lib.gbm_device_get_fd
gbm_device_get_fd.restype = c_int
gbm_device_get_fd.argtypes = [POINTER(struct_gbm_device)]

# /usr/include/gbm.h:259
gbm_device_get_backend_name = _lib.gbm_device_get_backend_name
gbm_device_get_backend_name.restype = c_char_p
gbm_device_get_backend_name.argtypes = [POINTER(struct_gbm_device)]

# /usr/include/gbm.h:263
gbm_device_is_format_supported = _lib.gbm_device_is_format_supported
gbm_device_is_format_supported.restype = c_int
gbm_device_is_format_supported.argtypes = [POINTER(struct_gbm_device), c_uint32, c_uint32]

# /usr/include/gbm.h:267
gbm_device_get_format_modifier_plane_count = _lib.gbm_device_get_format_modifier_plane_count
gbm_device_get_format_modifier_plane_count.restype = c_int
gbm_device_get_format_modifier_plane_count.argtypes = [POINTER(struct_gbm_device), c_uint32, c_uint64]

# /usr/include/gbm.h:272
gbm_device_destroy = _lib.gbm_device_destroy
gbm_device_destroy.restype = None
gbm_device_destroy.argtypes = [POINTER(struct_gbm_device)]

# /usr/include/gbm.h:274
gbm_create_device = _lib.gbm_create_device
gbm_create_device.restype = POINTER(struct_gbm_device)
gbm_create_device.argtypes = [c_int]


# /usr/include/gbm.h:277
gbm_bo_create = _lib.gbm_bo_create
gbm_bo_create.restype = POINTER(struct_gbm_bo)
gbm_bo_create.argtypes = [POINTER(struct_gbm_device), c_uint32, c_uint32, c_uint32, c_uint32]

# /usr/include/gbm.h:282
gbm_bo_create_with_modifiers = _lib.gbm_bo_create_with_modifiers
gbm_bo_create_with_modifiers.restype = POINTER(struct_gbm_bo)
gbm_bo_create_with_modifiers.argtypes = [POINTER(struct_gbm_device), c_uint32, c_uint32, c_uint32, POINTER(c_uint64),
                                         c_uint]

# /usr/include/gbm.h:289
gbm_bo_create_with_modifiers2 = _lib.gbm_bo_create_with_modifiers2
gbm_bo_create_with_modifiers2.restype = POINTER(struct_gbm_bo)
gbm_bo_create_with_modifiers2.argtypes = [POINTER(struct_gbm_device), c_uint32, c_uint32, c_uint32, POINTER(c_uint64),
                                          c_uint, c_uint32]

GBM_BO_IMPORT_WL_BUFFER = 21761  # /usr/include/gbm.h:297
GBM_BO_IMPORT_EGL_IMAGE = 21762  # /usr/include/gbm.h:298
GBM_BO_IMPORT_FD = 21763  # /usr/include/gbm.h:299
GBM_BO_IMPORT_FD_MODIFIER = 21764  # /usr/include/gbm.h:300
GBM_MAX_PLANES = 4  # /usr/include/gbm.h:310

# /usr/include/gbm.h:323
gbm_bo_import = _lib.gbm_bo_import
gbm_bo_import.restype = POINTER(struct_gbm_bo)
gbm_bo_import.argtypes = [POINTER(struct_gbm_device), c_uint32, POINTER(None), c_uint32]

# /usr/include/gbm.h:354
gbm_bo_map = _lib.gbm_bo_map
gbm_bo_map.restype = POINTER(c_void)
gbm_bo_map.argtypes = [POINTER(struct_gbm_bo), c_uint32, c_uint32, c_uint32, c_uint32, c_uint32, POINTER(c_uint32),
                       POINTER(POINTER(None))]

# /usr/include/gbm.h:360
gbm_bo_unmap = _lib.gbm_bo_unmap
gbm_bo_unmap.restype = None
gbm_bo_unmap.argtypes = [POINTER(struct_gbm_bo), POINTER(None)]

# /usr/include/gbm.h:363
gbm_bo_get_width = _lib.gbm_bo_get_width
gbm_bo_get_width.restype = c_uint32
gbm_bo_get_width.argtypes = [POINTER(struct_gbm_bo)]

# /usr/include/gbm.h:366
gbm_bo_get_height = _lib.gbm_bo_get_height
gbm_bo_get_height.restype = c_uint32
gbm_bo_get_height.argtypes = [POINTER(struct_gbm_bo)]

# /usr/include/gbm.h:369
gbm_bo_get_stride = _lib.gbm_bo_get_stride
gbm_bo_get_stride.restype = c_uint32
gbm_bo_get_stride.argtypes = [POINTER(struct_gbm_bo)]

# /usr/include/gbm.h:372
gbm_bo_get_stride_for_plane = _lib.gbm_bo_get_stride_for_plane
gbm_bo_get_stride_for_plane.restype = c_uint32
gbm_bo_get_stride_for_plane.argtypes = [POINTER(struct_gbm_bo), c_int]

# /usr/include/gbm.h:375
gbm_bo_get_format = _lib.gbm_bo_get_format
gbm_bo_get_format.restype = c_uint32
gbm_bo_get_format.argtypes = [POINTER(struct_gbm_bo)]

# /usr/include/gbm.h:378
gbm_bo_get_bpp = _lib.gbm_bo_get_bpp
gbm_bo_get_bpp.restype = c_uint32
gbm_bo_get_bpp.argtypes = [POINTER(struct_gbm_bo)]

# /usr/include/gbm.h:381
gbm_bo_get_offset = _lib.gbm_bo_get_offset
gbm_bo_get_offset.restype = c_uint32
gbm_bo_get_offset.argtypes = [POINTER(struct_gbm_bo), c_int]

# /usr/include/gbm.h:383
gbm_bo_get_device = _lib.gbm_bo_get_device
gbm_bo_get_device.restype = POINTER(struct_gbm_device)
gbm_bo_get_device.argtypes = [POINTER(struct_gbm_bo)]

# /usr/include/gbm.h:387
gbm_bo_get_handle = _lib.gbm_bo_get_handle
gbm_bo_get_handle.restype = struct_gbm_bo_handle
gbm_bo_get_handle.argtypes = [POINTER(struct_gbm_bo)]

# /usr/include/gbm.h:390
gbm_bo_get_fd = _lib.gbm_bo_get_fd
gbm_bo_get_fd.restype = c_int
gbm_bo_get_fd.argtypes = [POINTER(struct_gbm_bo)]

# /usr/include/gbm.h:393
gbm_bo_get_modifier = _lib.gbm_bo_get_modifier
gbm_bo_get_modifier.restype = c_uint64
gbm_bo_get_modifier.argtypes = [POINTER(struct_gbm_bo)]

# /usr/include/gbm.h:396
gbm_bo_get_plane_count = _lib.gbm_bo_get_plane_count
gbm_bo_get_plane_count.restype = c_int
gbm_bo_get_plane_count.argtypes = [POINTER(struct_gbm_bo)]

# /usr/include/gbm.h:399
gbm_bo_get_handle_for_plane = _lib.gbm_bo_get_handle_for_plane
gbm_bo_get_handle_for_plane.restype = struct_gbm_bo_handle
gbm_bo_get_handle_for_plane.argtypes = [POINTER(struct_gbm_bo), c_int]

# /usr/include/gbm.h:402
gbm_bo_get_fd_for_plane = _lib.gbm_bo_get_fd_for_plane
gbm_bo_get_fd_for_plane.restype = c_int
gbm_bo_get_fd_for_plane.argtypes = [POINTER(struct_gbm_bo), c_int]

# /usr/include/gbm.h:405
gbm_bo_write = _lib.gbm_bo_write
gbm_bo_write.restype = c_int
gbm_bo_write.argtypes = [POINTER(struct_gbm_bo), POINTER(None), c_size_t]

# /usr/include/gbm.h:408
gbm_bo_set_user_data = _lib.gbm_bo_set_user_data
gbm_bo_set_user_data.restype = None
gbm_bo_set_user_data.argtypes = [POINTER(struct_gbm_bo), POINTER(None),
                                 CFUNCTYPE(None, POINTER(struct_gbm_bo), POINTER(None))]

# /usr/include/gbm.h:411
gbm_bo_get_user_data = _lib.gbm_bo_get_user_data
gbm_bo_get_user_data.restype = POINTER(c_void)
gbm_bo_get_user_data.argtypes = [POINTER(struct_gbm_bo)]

# /usr/include/gbm.h:415
gbm_bo_destroy = _lib.gbm_bo_destroy
gbm_bo_destroy.restype = None
gbm_bo_destroy.argtypes = [POINTER(struct_gbm_bo)]

# /usr/include/gbm.h:417
gbm_surface_create = _lib.gbm_surface_create
gbm_surface_create.restype = POINTER(struct_gbm_surface)
gbm_surface_create.argtypes = [POINTER(struct_gbm_device), c_uint32, c_uint32, c_uint32, c_uint32]

# /usr/include/gbm.h:422
gbm_surface_create_with_modifiers = _lib.gbm_surface_create_with_modifiers
gbm_surface_create_with_modifiers.restype = POINTER(struct_gbm_surface)
gbm_surface_create_with_modifiers.argtypes = [POINTER(struct_gbm_device), c_uint32, c_uint32, c_uint32,
                                              POINTER(c_uint64), c_uint]

# /usr/include/gbm.h:429
gbm_surface_create_with_modifiers2 = _lib.gbm_surface_create_with_modifiers2
gbm_surface_create_with_modifiers2.restype = POINTER(struct_gbm_surface)
gbm_surface_create_with_modifiers2.argtypes = [POINTER(struct_gbm_device), c_uint32, c_uint32, c_uint32,
                                               POINTER(c_uint64), c_uint, c_uint32]

# /usr/include/gbm.h:437
gbm_surface_lock_front_buffer = _lib.gbm_surface_lock_front_buffer
gbm_surface_lock_front_buffer.restype = POINTER(struct_gbm_bo)
gbm_surface_lock_front_buffer.argtypes = [POINTER(struct_gbm_surface)]

# /usr/include/gbm.h:441
gbm_surface_release_buffer = _lib.gbm_surface_release_buffer
gbm_surface_release_buffer.restype = None
gbm_surface_release_buffer.argtypes = [POINTER(struct_gbm_surface), POINTER(struct_gbm_bo)]

# /usr/include/gbm.h:444
gbm_surface_has_free_buffers = _lib.gbm_surface_has_free_buffers
gbm_surface_has_free_buffers.restype = c_int
gbm_surface_has_free_buffers.argtypes = [POINTER(struct_gbm_surface)]

# /usr/include/gbm.h:447
gbm_surface_destroy = _lib.gbm_surface_destroy
gbm_surface_destroy.restype = None
gbm_surface_destroy.argtypes = [POINTER(struct_gbm_surface)]

# /usr/include/gbm.h:449
gbm_format_get_name = _lib.gbm_format_get_name
gbm_format_get_name.restype = c_char_p
gbm_format_get_name.argtypes = [c_uint32, POINTER(struct_gbm_format_name_desc)]



#!/usr/bin/env python
# ----------------------------------------------------------------------------
# pyglet
# Copyright (c) 2006-2008 Alex Holkner
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#
#  * Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
#  * Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in
#    the documentation and/or other materials provided with the
#    distribution.
#  * Neither the name of pyglet nor the names of its
#    contributors may be used to endorse or promote products
#    derived from this software without specific prior written
#    permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
# COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
# ----------------------------------------------------------------------------
# $Id$

import math

import pyglet

# # Default to OpenAL if available:
# pyglet.options['audio'] = 'openal', 'pulse', 'directsound', 'silent'

from pyglet.gl import glClearColor
from pyglet.graphics import Group
from pyglet.math import Mat4, Vec3

import reader

pyglet.resource.path.append('res')
pyglet.resource.reindex()


def orientation_angle(orientation):
    return math.atan2(orientation[2], orientation[0])


# Shape hack functions
def _update_line(line, x, y, x2, y2):
    line._x = x
    line._y = y
    line._x2 = x2
    line._y2 = y2
    line._update_translation()
    line._update_vertices()


def _bordered_rect_alpha(rect, inner_alpha, border_alpha):
    rect._rgba = rect._rgba[:3] + (inner_alpha,)
    rect._border_rgba = rect._border_rgba[:3] + (border_alpha,)
    rect._update_color()


class Handle:
    tip = ''

    def __init__(self, window, player):
        self.win = window
        self.player = player

    def hit_test(self, x, y, z):
        dx, dy, dz = (a - b for a, b in zip(self.pos(), (x, y, z)))
        if dx * dx + dy * dy + dz * dz < self.radius * self.radius:
            return -dx, -dy, -dz

    def pos(self):
        raise NotImplementedError

    def delete(self):
        pass

    def update_shapes(self):
        pass

    def begin_drag(self, offset):
        self.offset = offset
        return self

    def on_mouse_press(self, x, y, button, modifiers):
        self.win.remove_handlers(self)

    def on_mouse_release(self, x, y, button, modifiers):
        self.win.remove_handlers(self)


class LabelHandle(Handle):
    def __init__(self, window, player):
        super().__init__(window, player)
        txt = getattr(player, 'label', '')
        self.text = pyglet.text.Label(txt, font_size=10, color=(0, 0, 0, 255),
                                      anchor_y='top', anchor_x='center', batch=window.text_batch)

    def delete(self):
        self.text.delete()

    def update_shapes(self):
        x, z = self.win.player_transform(self.player)
        self.text.position = x, z - 5, 0

    def hit_test(self, x, y, z):
        return None


class PositionHandle(Handle):
    tip = 'position'
    radius = .3

    def __init__(self, window, player):
        super().__init__(window, player)
        self._triangle = pyglet.shapes.Triangle(
            0, self.radius,
            -self.radius * math.sqrt(3.0) / 2.0, -self.radius / 2.0,
            self.radius * math.sqrt(3.0) / 2.0, -self.radius / 2.0,
            (255, 0, 0, 255),
            window.handle_batch,
            window.hgrp_position)
        self._triangle.anchor_position = 0.0, -self.radius * 0.75

    def update_shapes(self):
        self._triangle.position = self.player.position[0], self.player.position[2]

    def delete(self):
        self._triangle.delete()

    def pos(self):
        return self.player.position

    def on_mouse_drag(self, x, y, dx, dy, buttons, modifiers):
        pos = self.win.mouse_transform(x, y)
        self.player.position = \
            (pos[0] - self.offset[0],
             pos[1] - self.offset[1],
             pos[2] - self.offset[2])


class OrientationHandle(Handle):
    radius = .1
    length = 1.5

    def __init__(self, window, player):
        super().__init__(window, player)
        self._line = pyglet.shapes.Line(0, 0, 0, 0,
                                        color=(74, 74, 74),
                                        batch=window.handle_batch,
                                        group=Group(0, window.hgrp_orientation))
        self._disc = pyglet.shapes.Circle(0, 0, self.radius,
                                          color=(255, 255, 0, 255),
                                          batch=window.handle_batch,
                                          group=Group(1, window.hgrp_orientation))

    def pos(self):
        x, _, z = self.player.position
        direction = self.get_orientation()
        sz = math.sqrt(direction[0] ** 2 + direction[1] ** 2 + direction[2] ** 2) or 1
        if sz != 0:
            x += direction[0] / sz * self.length
            z += direction[2] / sz * self.length
        return x, 0, z

    def get_orientation(self):
        return NotImplementedError()

    def update_shapes(self):
        px, _, py = self.player.position
        x, _, y = self.pos()

        self._line._thickness = 1.0 / self.win.zoom
        _update_line(self._line, x, y, px, py)

        self._disc.position = x, y

    def delete(self):
        self._line.delete()
        self._disc.delete()

    def on_mouse_drag(self, x, y, dx, dy, buttons, modifiers):
        px, py, pz = self.player.position
        hx, hy, hz = self.win.mouse_transform(x, y)
        self.set_orientation(
            (hx - self.offset[0] - px,
             hy - self.offset[1] - py,
             hz - self.offset[2] - pz))


class ConeOrientationHandle(OrientationHandle):
    tip = 'cone_orientation'

    def get_orientation(self):
        return self.player.cone_orientation

    def set_orientation(self, orientation):
        self.player.cone_orientation = orientation


class ForwardOrientationHandle(OrientationHandle):
    tip = 'forward_orientation'

    def get_orientation(self):
        return self.player.forward_orientation

    def set_orientation(self, orientation):
        self.player.forward_orientation = orientation


class ConeAngleHandle(Handle):
    length = .1
    radius = .1

    def __init__(self, window, player):
        super().__init__(window, player)
        self._sector = pyglet.shapes.Sector(0, 0, self.length, color=self.fill_color, batch=window.handle_batch)
        self._handle = pyglet.shapes.Circle(0, 0, self.radius, color=self.color, batch=window.handle_batch)

    def pos(self):
        px, py, pz = self.player.position
        angle = orientation_angle(self.player.cone_orientation)
        angle += self.get_angle() * math.pi / 180. / 2
        x = math.cos(angle) * self.length
        z = math.sin(angle) * self.length
        return px + x, py, pz + z

    def get_angle(self):
        raise NotImplementedError

    def update_shapes(self):
        px, _, py = self.player.position
        angle = orientation_angle(self.player.cone_orientation)
        a = math.radians(self.get_angle())

        self._sector.angle = self.get_angle()
        self._sector.rotation = math.degrees((-angle / 2.0) + (a / 4.0))
        self._sector.position = px, py

        x, _, y = self.pos()
        self._handle.radius = self.radius
        self._handle.position = x, y

    def delete(self):
        self._sector.delete()
        self._handle.delete()

    def on_mouse_drag(self, x, y, dx, dy, buttons, modifiers):
        px, py, pz = self.player.position
        hx, hy, hz = self.win.mouse_transform(x, y)
        angle = orientation_angle(self.player.cone_orientation)
        hangle = orientation_angle((hx - px, hy - py, hz - pz))
        if hangle < angle:
            hangle += math.pi * 2
        res = min(max((hangle - angle) * 2, 0), math.pi * 2)
        self.set_angle(res * 180. / math.pi)


class ConeInnerAngleHandle(ConeAngleHandle):
    tip = 'cone_inner_angle'
    length = 1.
    color = (51, 204, 51, 255)
    fill_color = (0, 255, 0, 26)

    def get_angle(self):
        return self.player.cone_inner_angle

    def set_angle(self, angle):
        self.player.cone_inner_angle = angle


class ConeOuterAngleHandle(ConeAngleHandle):
    tip = 'cone_outer_angle'
    length = 1.2
    color = (51, 51, 204, 255)
    fill_color = (0, 0, 255, 26)

    def get_angle(self):
        return self.player.cone_outer_angle

    def set_angle(self, angle):
        self.player.cone_outer_angle = angle


class MoreHandle(Handle):
    tip = 'More...'
    radius = .2

    open = False
    open_width = 1.5
    open_height = 1.5

    def __init__(self, window, player):
        super().__init__(window, player)
        self._box = pyglet.shapes.BorderedRectangle(0, 0, self.open_width, self.open_height, 1,
                                                    (255, 255, 255), (0, 0, 0),
                                                    window.handle_batch,
                                                    Group(0, window.hgrp_label_more_collapsed))
        # The BorderedRectangle vehemently defends itself from having differing alpha values,
        # but hack them in case we want them
        _bordered_rect_alpha(self._box, 201, 255)

        self._circ = pyglet.shapes.Circle(0, 0, self.radius,
                                          batch=window.handle_batch,
                                          group=Group(0, window.hgrp_label_more_collapsed))
        self._outline = pyglet.shapes.Arc(0, 0, self.radius,
                                          color=(0, 0, 0, 255),
                                          batch=window.handle_batch,
                                          group=Group(1, window.hgrp_label_more_collapsed))
        self._line0 = pyglet.shapes.Line(0, 0, 0, 0,
                                         color=(0, 0, 0, 255),
                                         batch=window.handle_batch,
                                         group=Group(2, window.hgrp_label_more_collapsed))
        self._line1 = pyglet.shapes.Line(0, 0, 0, 0,
                                         color=(0, 0, 0, 255),
                                         batch=window.handle_batch,
                                         group=Group(3, window.hgrp_label_more_collapsed))

    def pos(self):
        x, y, z = self.player.position
        return x + 1, y, z + 1

    def get_angle(self):
        raise NotImplementedError

    def update_shapes(self):
        self._box.visible = self.open
        self._circ.visible = not self.open
        self._outline.visible = not self.open
        self._line0.visible = not self.open
        self._line1.visible = not self.open

        x, _, z = self.pos()
        if self.open:
            x -= 0.2
            z += 0.2
            self._box.border = 1.0 / self.win.zoom
            self._box.position = x, z - self.open_height
        else:
            self._circ.position = x, z
            self._outline.position = x, z
            self._outline.thickness = 1.0 / self.win.zoom
            r = self.radius - 0.1
            self._line0._thickness = 1.0 / self.win.zoom
            _update_line(self._line0, x - r, z, x + r, z)
            self._line1._thickness = 1.0 / self.win.zoom
            _update_line(self._line1, x, z - r, x, z + r)

    def delete(self):
        self._box.delete()
        self._circ.delete()
        self._outline.delete()
        self._line0.delete()
        self._line1.delete()

    def begin_drag(self, offset):
        self.open = True
        self.win.set_more_player_handles(self.player)
        return self

    def on_mouse_press(self, x, y, button, modifiers):
        x, y, z = self.win.mouse_transform(x, y)
        for handle in self.win.more_handles:
            if handle.hit_test(x, y, z):
                return
        self.win.set_more_player_handles(None)
        self.win.remove_handlers(self)
        self.open = False

    def on_mouse_release(self, x, y, button, modifiers):
        pass


class SliderHandle(Handle):
    length = 1.
    width = .05
    radius = .1

    def __init__(self, window, player, x, z):
        super().__init__(window, player)
        self._groove = pyglet.shapes.Rectangle(0, 0, self.length, self.width,
                                               color=(127, 127, 127, 255),
                                               batch=window.handle_batch,
                                               group=Group(0, window.hgrp_more_expanded))
        self._thumb = pyglet.shapes.Circle(0, 0, self.radius,
                                           color=(51, 51, 51, 255),
                                           batch=window.handle_batch,
                                           group=Group(1, window.hgrp_more_expanded))
        self.x = x
        self.z = z

    def pos(self):
        x, y, z = self.player.position
        x += self.x + self.get_value() * self.length
        z += self.z
        return x, y, z

    def update_shapes(self):
        gx = self.x + self.player.position[0]
        gz = self.z + self.player.position[2]
        self._groove.position = gx, gz

        x, _, z = self.pos()
        self._thumb.position = x, z

    def delete(self):
        self._groove.delete()
        self._thumb.delete()

    def on_mouse_drag(self, x, y, dx, dy, buttons, modifiers):
        px, py, pz = self.player.position
        hx, hy, hz = self.win.mouse_transform(x, y)
        value = float(hx - px - self.x) / self.length
        value = min(max(value, 0), 1)
        self.set_value(value)


class VolumeHandle(SliderHandle):
    tip = 'volume'

    def __init__(self, window, player):
        super().__init__(window, player, 1, .9)

    def get_value(self):
        return self.player.volume

    def set_value(self, value):
        self.player.volume = value


class ListenerVolumeHandle(SliderHandle):
    tip = 'volume'

    def __init__(self, window, player):
        super().__init__(window, player, -.5, -1)

    def get_value(self):
        return self.player.volume

    def set_value(self, value):
        self.player.volume = value


class MinDistanceHandle(SliderHandle):
    tip = 'min_distance'

    def __init__(self, window, player):
        super().__init__(window, player, 1, .6)

    def get_value(self):
        return self.player.min_distance / 5.

    def set_value(self, value):
        self.player.min_distance = value * 5.


class MaxDistanceHandle(SliderHandle):
    tip = 'max_distance'

    def __init__(self, window, player):
        super().__init__(window, player, 1, .3)

    def get_value(self):
        return min(self.player.max_distance / 5., 1.0)

    def set_value(self, value):
        self.player.max_distance = value * 5.


class ConeOuterGainHandle(SliderHandle):
    tip = 'cone_outer_gain'

    def __init__(self, window, player):
        super().__init__(window, player, 1, 0)

    def get_value(self):
        return self.player.cone_outer_gain

    def set_value(self, value):
        self.player.cone_outer_gain = value


class SoundSpaceWindow(pyglet.window.Window):
    def __init__(self, **kwargs):
        kwargs.setdefault("caption", "Sound Space")
        kwargs.setdefault("resizable", True)
        super().__init__(**kwargs)

        # pixels per unit
        self.zoom = 40
        self.tx = self.width/2
        self.ty = self.height/2

        # The handle batch is drawn with a greatly zoomed view matrix
        self._grid_batch = pyglet.graphics.Batch()
        self.handle_batch = pyglet.graphics.Batch()
        self.text_batch = pyglet.graphics.Batch()

        # Various handle groups for proper rendering order
        self.hgrp_position = Group(0)
        self.hgrp_orientation = Group(1)
        self.hgrp_cone_inner = Group(2)
        self.hgrp_cone_outer = Group(3)
        self.hgrp_label_more_collapsed = Group(4)
        self.hgrp_more_expanded = Group(5)

        self._grid = []
        self._refresh_grid()
        self.players = []
        self.handles = []
        self.more_handles = []

        listener = pyglet.media.get_audio_driver().get_listener()
        self.handles.append(PositionHandle(self, listener))
        self.handles.append(ForwardOrientationHandle(self, listener))
        self.handles.append(ListenerVolumeHandle(self, listener))
        self.handles.append(LabelHandle(self, listener))

        self.tip = pyglet.text.Label('', font_size=10, color=(0, 0, 0, 255),
                                     anchor_y='top', anchor_x='center')
        self.tip_player = None

    def add_player(self, player):
        self.players.append(player)
        self.handles.append(PositionHandle(self, player))
        self.handles.append(ConeOrientationHandle(self, player))
        self.handles.append(ConeInnerAngleHandle(self, player))
        self.handles.append(ConeOuterAngleHandle(self, player))
        self.handles.append(LabelHandle(self, player))
        self.handles.append(MoreHandle(self, player))

    def set_more_player_handles(self, player):
        if player:
            self.more_handles = [
                VolumeHandle(self, player),
                MinDistanceHandle(self, player),
                MaxDistanceHandle(self, player),
                ConeOuterGainHandle(self, player),
            ]
        else:
            for h in self.more_handles:
                h.delete()
            self.more_handles = []

    def _refresh_grid(self):
        for line in self._grid:
            line.delete()
        self._grid.clear()

        for x in range(0, self.width, self.zoom):
            self._grid.append(pyglet.shapes.Line(x, 0, x, self.height, batch=self._grid_batch))
        for y in range(0, self.height, self.zoom):
            self._grid.append(pyglet.shapes.Line(0, y, self.width, y, batch=self._grid_batch))

    def mouse_transform(self, x, y):
        return (float(x - self.tx) / self.zoom,
                0,
                float(y - self.ty) / self.zoom)

    def player_transform(self, player):
        return (player.position[0] * self.zoom + self.tx,
                player.position[2] * self.zoom + self.ty)

    def hit_test(self, mouse_x, mouse_y):
        x, y, z = self.mouse_transform(mouse_x, mouse_y)
        for handle in self.more_handles[::-1] + self.handles[::-1]:
            offset = handle.hit_test(x, y, z)
            if offset:
                return handle, offset
        return None, None

    def on_draw(self):
        glClearColor(.8, .8, .8, 1)
        self.clear()

        for handle in self.handles + self.more_handles:
            handle.update_shapes()

        self.view = Mat4()
        self._grid_batch.draw()

        self.view = Mat4().translate(Vec3(self.tx, self.ty, 0)).scale(Vec3(self.zoom, self.zoom, 0))
        self.handle_batch.draw()

        self.view = Mat4()
        self.text_batch.draw()

        if self.tip_player:
            player_pos = self.player_transform(self.tip_player)
            self.tip.x = player_pos[0]
            self.tip.y = player_pos[1] - 15
            self.tip.draw()

    def on_mouse_scroll(self, x, y, dx, dy):
        self.zoom += dy * 10
        self.zoom = min(max(self.zoom, 10), 100)
        self._refresh_grid()

    def on_mouse_press(self, x, y, button, modifiers):
        handle, offset = self.hit_test(x, y)
        if handle:
            self.push_handlers(handle.begin_drag(offset))
        else:
            self.push_handlers(
                on_mouse_press=self._pan_on_mouse_press_or_release,
                on_mouse_release=self._pan_on_mouse_press_or_release,
                on_mouse_drag=self._pan_on_mouse_drag,
            )

    def _pan_on_mouse_press_or_release(self, _x, _y, _b, _m):
        self.remove_handlers(
            on_mouse_press=self._pan_on_mouse_press_or_release,
            on_mouse_release=self._pan_on_mouse_press_or_release,
            on_mouse_drag=self._pan_on_mouse_drag,
        )

    def _pan_on_mouse_drag(self, x, y, dx, dy, buttons, modifiers):
        self.tx += dx
        self.ty += dy

    def on_mouse_motion(self, x, y, dx, dy):
        handle, offset = self.hit_test(x, y)
        if handle:
            self.tip.text = handle.tip
            pos = self.player_transform(handle.player)
            self.tip_player = handle.player
        else:
            self.tip.text = ''


if __name__ == '__main__':
    # We swap Y and Z, moving to left-handed system
    listener = pyglet.media.get_audio_driver().get_listener()
    listener.up_orientation = (0, -1, 0)

    # Start facing up (er, forwards)
    listener.forward_orientation = (0, 0, 1)

    listener.label = 'Listener'

    w = SoundSpaceWindow()
    r = reader.SpaceReader(w)
    r.read(pyglet.resource.file('space.txt', 'r'))
    player_group = pyglet.media.PlayerGroup(w.players)
    player_group.play()

    pyglet.app.run()



"""Wrapper for include/libavcodec/avcodec.h
"""

from ctypes import c_int, c_uint16, c_int64, c_uint32, c_uint64, c_size_t
from ctypes import c_uint8, c_uint, c_float, c_char_p
from ctypes import c_void_p, POINTER, CFUNCTYPE, Structure

import pyglet.lib
from pyglet.util import debug_print
from . import compat
from . import libavutil
from .libavutil import AVChannelLayout, AVDictionary

_debug = debug_print('debug_media')

avcodec = pyglet.lib.load_library(
    'avcodec',
    win32=('avcodec-61', 'avcodec-60', 'avcodec-59', 'avcodec-58'),
    darwin=('avcodec.61', 'avcodec.60', 'avcodec.59', 'avcodec.58')
)

avcodec.avcodec_version.restype = c_int

avcodec_version = avcodec.avcodec_version() >> 16

compat.set_version('avcodec', avcodec_version)


FF_INPUT_BUFFER_PADDING_SIZE = 32


class AVPacketSideData(Structure):
    _fields_ = [
        ('data', POINTER(c_uint8)),
        ('size', c_int),
        ('type', c_int)
    ]


AVBufferRef = libavutil.AVBufferRef

AVRational = libavutil.AVRational


class AVPacket(Structure):
    pass

AVPacket_Fields = [
        ('buf', POINTER(AVBufferRef)),
        ('pts', c_int64),
        ('dts', c_int64),
        ('data', POINTER(c_uint8)),
        ('size', c_int),
        ('stream_index', c_int),
        ('flags', c_int),
        ('side_data', POINTER(AVPacketSideData)),
        ('side_data_elems', c_int),
        ('duration', c_int64),
        ('pos', c_int64),
        ('opaque', c_void_p),  # 5.x+
        ('opaque_ref', c_void_p),  # 5.x+
        ('time_base', AVRational),  # 5.x+
        ('convergence_duration', c_int64)  # 4.x only
    ]


compat.add_version_changes('avcodec', 58, AVPacket, AVPacket_Fields,
                           removals=('opaque', 'opaque_ref', 'time_base'))

for compat_ver in (59, 60, 61):
    compat.add_version_changes('avcodec', compat_ver, AVPacket, AVPacket_Fields,
                               removals=('convergence_duration',))

class AVCodecParserContext(Structure):
    pass


class AVCodecParameters(Structure):
    pass

AVCodecParameters_Fields = [
    ('codec_type', c_int),
    ('codec_id', c_int),
    ('codec_tag', c_uint32),
    ('extradata', POINTER(c_uint8)),
    ('extradata_size', c_int),
    ('coded_side_data', POINTER(AVPacketSideData)),  # added in 61
    ('nb_coded_side_data', c_int),  # added in 61
    ('format', c_int),
    ('bit_rate', c_int64),
    ('bits_per_coded_sample', c_int),
    ('bits_per_raw_sample', c_int),
    ('profile', c_int),
    ('level', c_int),
    ('width', c_int),
    ('height', c_int),
    ('sample_aspect_ratio', AVRational),
    ('framerate', AVRational),  # added in 61
    ('field_order', c_int),
    ('color_range', c_int),
    ('color_primaries', c_int),
    ('color_trc', c_int),
    ('color_space', c_int),
    ('chroma_location', c_int),
    ('video_delay', c_int),
    ('ch_layout', AVChannelLayout),  # added in 61
    ('channel_layout', c_uint64),  # deprecated as of 59. removed in 61
    ('channels', c_int),  # deprecated as of 59. removed in 61
    ('sample_rate', c_int),
    ('block_align', c_int),
    ('frame_size', c_int),
    ('initial_padding', c_int),
    ('trailing_padding', c_int),
    ('seek_preroll', c_int),
]

for compat_ver in (58, 59, 60):
    compat.add_version_changes('avcodec', compat_ver, AVCodecParameters, AVCodecParameters_Fields,
                               removals=('coded_side_data', 'nb_coded_side_data', 'ch_layout', 'framerate'))

compat.add_version_changes('avcodec', 61, AVCodecParameters, AVCodecParameters_Fields,
                           removals=('channel_layout', 'channels'))


class AVProfile(Structure):
    _fields_ = [
        ('profile', c_int),
        ('name', c_char_p),
    ]


class AVCodecDescriptor(Structure):
    _fields_ = [
        ('id', c_int),
        ('type', c_int),
        ('name', c_char_p),
        ('long_name', c_char_p),
        ('props', c_int),
        ('mime_types', c_char_p),
        ('profiles', POINTER(AVProfile))
    ]


class AVCodecInternal(Structure):
    pass


class AVCodec(Structure):
    _fields_ = [
        ('name', c_char_p),
        ('long_name', c_char_p),
        ('type', c_int),
        ('id', c_int),
        ('capabilities', c_int),
        ('supported_framerates', POINTER(AVRational)),
        ('pix_fmts', POINTER(c_int)),
        ('supported_samplerates', POINTER(c_int)),
        ('sample_fmts', POINTER(c_int)),
        ('channel_layouts', POINTER(c_uint64)),
        ('max_lowres', c_uint8),
        # And more...
    ]


class AVCodecContext(Structure):
    pass


class RcOverride(Structure):
    pass


class AVHWAccel(Structure):
    pass

AVFrameSideDataType = c_int

class AVFrameSideData(Structure):
    _fields_ = [
        ("type", AVFrameSideDataType),
        ("data", POINTER(c_uint8)),
        ("size", c_size_t),
        ("metadata", POINTER(AVDictionary)),
        ("buf", POINTER(AVBufferRef))
    ]

AVClass = libavutil.AVClass
AVFrame = libavutil.AVFrame
AV_NUM_DATA_POINTERS = libavutil.AV_NUM_DATA_POINTERS

# Significant deprecation and re-ordering of the entire structure makes it unmanagable to
# track of all the changes via compat module. Re-define the structure and compat the new one going forward.
if avcodec_version >= 61:
    AVCodecContext_Fields = [
        # Basic fields
        ("av_class", POINTER(AVClass)),
        ("log_level_offset", c_int),

        # Codec fields
        ("codec_type", c_int),  # enum AVMediaType
        ("codec", POINTER(AVCodec)),
        ("codec_id", c_int),  # enum AVCodecID
        ("codec_tag", c_uint),

        ("priv_data", c_void_p),
        ("internal", POINTER(AVCodecInternal)),
        ("opaque", c_void_p),
        ("bit_rate", c_int64),
        ("flags", c_int),
        ("flags2", c_int),
        ("extradata", POINTER(c_uint8)),
        ("extradata_size", c_int),

        # Timebase
        ("time_base", AVRational),
        ("pkt_timebase", AVRational),
        ("framerate", AVRational),

        # Video fields
        ("ticks_per_frame", c_int),  # Deprecated in 61.
        ("delay", c_int),
        ("width", c_int),
        ("height", c_int),
        ("coded_width", c_int),
        ("coded_height", c_int),
        ("sample_aspect_ratio", AVRational),  # AVRational
        ("pix_fmt", c_int),  # enum AVPixelFormat
        ("sw_pix_fmt", c_int),  # enum AVPixelFormat
        ("color_primaries", c_int),  # enum AVColorPrimaries
        ("color_trc", c_int),  # enum AVColorTransferCharacteristic
        ("colorspace", c_int),  # enum AVColorSpace
        ("color_range", c_int),  # enum AVColorRange
        ("chroma_sample_location", c_int),  # enum AVChromaLocation
        ("field_order", c_int),  # enum AVFieldOrder
        ("refs", c_int),
        ("has_b_frames", c_int),
        ("slice_flags", c_int),

        ("draw_horiz_band", CFUNCTYPE(None, POINTER(AVCodecContext), POINTER(AVFrame), POINTER(c_int), c_int, c_int, c_int)),
        ("get_format", CFUNCTYPE(c_int, POINTER(AVCodecContext), POINTER(c_int))),

        # Video encoding parameters
        ("max_b_frames", c_int),
        ("b_quant_factor", c_float),
        ("b_quant_offset", c_float),
        ("i_quant_factor", c_float),
        ("i_quant_offset", c_float),
        ("lumi_masking", c_float),
        ("temporal_cplx_masking", c_float),
        ("spatial_cplx_masking", c_float),
        ("p_masking", c_float),
        ("dark_masking", c_float),
        ("nsse_weight", c_int),
        ("me_cmp", c_int),
        ("me_sub_cmp", c_int),
        ("mb_cmp", c_int),
        ("ildct_cmp", c_int),
        ("dia_size", c_int),
        ("last_predictor_count", c_int),
        ("me_pre_cmp", c_int),
        ("pre_dia_size", c_int),
        ("me_subpel_quality", c_int),
        ("me_range", c_int),
        ("mb_decision", c_int),
        ("intra_matrix", POINTER(c_uint16)),
        ("inter_matrix", POINTER(c_uint16)),
        ("chroma_intra_matrix", POINTER(c_uint16)),
        ("intra_dc_precision", c_int),
        ("mb_lmin", c_int),
        ("mb_lmax", c_int),
        ("bidir_refine", c_int),
        ("keyint_min", c_int),
        ("gop_size", c_int),
        ("mv0_threshold", c_int),
        ("slices", c_int),

        # Audio fields
        ("sample_rate", c_int),
        ("sample_fmt", c_int),  # enum AVSampleFormat
        ("ch_layout", AVChannelLayout),  # AVChannelLayout
        ("frame_size", c_int),
        ("block_align", c_int),
        ("cutoff", c_int),
        ("audio_service_type", c_int),  # enum AVAudioServiceType
        ("request_sample_fmt", c_int),  # enum AVSampleFormat
        ("initial_padding", c_int),
        ("trailing_padding", c_int),
        ("seek_preroll", c_int),

        # Encoding parameters
        ("bit_rate_tolerance", c_int),
        ("global_quality", c_int),
        ("compression_level", c_int),
        ("qcompress", c_float),
        ("qblur", c_float),
        ("qmin", c_int),
        ("qmax", c_int),
        ("max_qdiff", c_int),
        ("rc_buffer_size", c_int),
        ("rc_override_count", c_int),
        ("rc_override", POINTER(RcOverride)),
        ("rc_max_rate", c_int64),
        ("rc_min_rate", c_int64),
        ("rc_max_available_vbv_use", c_float),
        ("rc_min_vbv_overflow_use", c_float),
        ("rc_initial_buffer_occupancy", c_int),
        ("trellis", c_int),
        ("stats_out", c_char_p),
        ("stats_in", c_char_p),
        ("workaround_bugs", c_int),
        ("strict_std_compliance", c_int),
        ("error_concealment", c_int),
        ("debug", c_int),
        ("err_recognition", c_int),
        ("hwaccel", POINTER(AVHWAccel)),
        ("hwaccel_context", c_void_p),
        ("hw_frames_ctx", POINTER(AVBufferRef)),
        ("hw_device_ctx", POINTER(AVBufferRef)),
        ("hwaccel_flags", c_int),
        ("extra_hw_frames", c_int),
        ("error", c_uint64 * AV_NUM_DATA_POINTERS),
        ("dct_algo", c_int),
        ("idct_algo", c_int),
        ("bits_per_coded_sample", c_int),
        ("bits_per_raw_sample", c_int),
        ("thread_count", c_int),
        ("thread_type", c_int),
        ("active_thread_type", c_int),
        ("execute",
         CFUNCTYPE(c_int, POINTER(AVCodecContext), CFUNCTYPE(c_int, POINTER(AVCodecContext), c_void_p), c_void_p,
                   POINTER(c_int), c_int, c_int)),
        ("execute2", CFUNCTYPE(c_int, POINTER(AVCodecContext),
                               CFUNCTYPE(c_int, POINTER(AVCodecContext), c_void_p, c_int, c_int), c_void_p,
                               POINTER(c_int), c_int)),
        ("profile", c_int),
        ("level", c_int),
        ("properties", c_uint),
        ("skip_loop_filter", c_int),  # enum AVDiscard
        ("skip_idct", c_int),  # enum AVDiscard
        ("skip_frame", c_int),  # enum AVDiscard
        ("skip_alpha", c_int),
        ("skip_top", c_int),
        ("skip_bottom", c_int),
        ("lowres", c_int),
        ("codec_descriptor", POINTER(AVCodecDescriptor)),
        ("sub_charenc", c_char_p),
        ("sub_charenc_mode", c_int),
        ("subtitle_header_size", c_int),
        ("subtitle_header", POINTER(c_uint8)),
        ("dump_separator", POINTER(c_uint8)),
        ("codec_whitelist", c_char_p),
        ("coded_side_data", POINTER(AVPacketSideData)),
        ("nb_coded_side_data", c_int),
        ("export_side_data", c_int),
        ("max_pixels", c_int64),
        ("apply_cropping", c_int),
        ("discard_damaged_percentage", c_int),
        ("max_samples", c_int64),
        ("get_encode_buffer", CFUNCTYPE(c_int, POINTER(AVCodecContext), POINTER(AVPacket), c_int)),
        ("frame_num", c_int64),
        ("side_data_prefer_packet", POINTER(c_int)),
        ("nb_side_data_prefer_packet", c_uint),
        ("decoded_side_data", POINTER(POINTER(AVFrameSideData))),
        ("nb_decoded_side_data", c_int),
    ]

    compat.add_version_changes('avcodec', 61, AVCodecContext, AVCodecContext_Fields, removals=None)

else:
    AVCodecContext_Fields = [
        ('av_class', POINTER(AVClass)),
        ('log_level_offset', c_int),
        ('codec_type', c_int),
        ('codec', POINTER(AVCodec)),
        ('codec_id', c_int),
        ('codec_tag', c_uint),
        ('priv_data', c_void_p),
        ('internal', POINTER(AVCodecInternal)),
        ('opaque', c_void_p),
        ('bit_rate', c_int64),
        ('bit_rate_tolerance', c_int),
        ('global_quality', c_int),
        ('compression_level', c_int),
        ('flags', c_int),
        ('flags2', c_int),
        ('extradata', POINTER(c_uint8)),
        ('extradata_size', c_int),
        ('time_base', AVRational),
        ('ticks_per_frame', c_int),  # deprecated. to be removed post 61.
        ('delay', c_int),
        ('width', c_int),
        ('height', c_int),
        ('coded_width', c_int),
        ('coded_height', c_int),
        ('gop_size', c_int),
        ('pix_fmt', c_int),
        ('draw_horiz_band', CFUNCTYPE(None,
                                      POINTER(AVCodecContext), POINTER(AVFrame),
                                      c_int * 8, c_int, c_int, c_int)),
        ('get_format', CFUNCTYPE(c_int, POINTER(AVCodecContext), POINTER(c_int))),
        ('max_b_frames', c_int),
        ('b_quant_factor', c_float),
        ('b_frame_strategy', c_int),  # Deprecated. Removed in 59.
        ('b_quant_offset', c_float),
        ('has_b_frames', c_int),
        ('mpeg_quant', c_int),  # Deprecated. Removed in 59.
        ('i_quant_factor', c_float),
        ('i_quant_offset', c_float),
        ('lumi_masking', c_float),
        ('temporal_cplx_masking', c_float),
        ('spatial_cplx_masking', c_float),
        ('p_masking', c_float),
        ('dark_masking', c_float),
        ('slice_count', c_int),
        ('prediction_method', c_int),  # Deprecated. Removed in 59.
        ('slice_offset', POINTER(c_int)),
        ('sample_aspect_ratio', AVRational),  # Moved in 61
        ('me_cmp', c_int),
        ('me_sub_cmp', c_int),
        ('mb_cmp', c_int),
        ('ildct_cmp', c_int),
        ('dia_size', c_int),
        ('last_predictor_count', c_int),
        ('pre_me', c_int),  # Deprecated. Removed in 59.
        ('me_pre_cmp', c_int),
        ('pre_dia_size', c_int),
        ('me_subpel_quality', c_int),
        ('me_range', c_int),
        ('slice_flags', c_int),
        ('mb_decision', c_int),
        ('intra_matrix', POINTER(c_uint16)),
        ('inter_matrix', POINTER(c_uint16)),
        ('scenechange_threshold', c_int),  # Deprecated. Removed in 59.
        ('noise_reduction', c_int),  # Deprecated. Removed in 59.
        ('intra_dc_precision', c_int),
        ('skip_top', c_int),
        ('skip_bottom', c_int),
        ('mb_lmin', c_int),
        ('mb_lmax', c_int),
        ('me_penalty_compensation', c_int),  # Deprecated. Removed in 59.
        ('bidir_refine', c_int),
        ('brd_scale', c_int),  # Deprecated. Removed in 59.
        ('keyint_min', c_int),
        ('refs', c_int),
        ('chromaoffset', c_int),  # Deprecated. Removed in 59.
        ('mv0_threshold', c_int),
        ('b_sensitivity', c_int),  # Deprecated. Removed in 59.
        ('color_primaries', c_int),
        ('color_trc', c_int),
        ('colorspace', c_int),
        ('color_range', c_int),
        ('chroma_sample_location', c_int),
        ('slices', c_int),
        ('field_order', c_int),
        ('sample_rate', c_int),
        ('channels', c_int),
        ('sample_fmt', c_int),
        ('frame_size', c_int),
        ('frame_number', c_int),
        ('block_align', c_int),
        ('cutoff', c_int),
        ('channel_layout', c_uint64),
        ('request_channel_layout', c_uint64),
        ('audio_service_type', c_int),
        ('request_sample_fmt', c_int),
        ('get_buffer2', CFUNCTYPE(c_int, POINTER(AVCodecContext), POINTER(AVFrame), c_int)),
        ('refcounted_frames', c_int),  # Deprecated. Removed in 59.
        ('qcompress', c_float),
        ('qblur', c_float),
        ('qmin', c_int),
        ('qmax', c_int),
        ('max_qdiff', c_int),
        ('rc_buffer_size', c_int),
        ('rc_override_count', c_int),
        ('rc_override', POINTER(RcOverride)),
        ('rc_max_rate', c_int64),
        ('rc_min_rate', c_int64),
        ('rc_max_available_vbv_use', c_float),
        ('rc_min_vbv_overflow_use', c_float),
        ('rc_initial_buffer_occupancy', c_int),
        ('coder_type', c_int),  # Deprecated. Removed in 59.
        ('context_model', c_int),  # Deprecated. Removed in 59.
        ('frame_skip_threshold', c_int),  # Deprecated. Removed in 59.
        ('frame_skip_factor', c_int),  # Deprecated. Removed in 59.
        ('frame_skip_exp', c_int),  # Deprecated. Removed in 59.
        ('frame_skip_cmp', c_int),  # Deprecated. Removed in 59.
        ('trellis', c_int),
        ('min_prediction_order', c_int),  # Deprecated. Removed in 59.
        ('max_prediction_order', c_int),  # Deprecated. Removed in 59.
        ('timecode_frame_start', c_int64),  # Deprecated. Removed in 59.
        ('rtp_callback', CFUNCTYPE(None,  # Deprecated. Removed in 59.
                                  POINTER(AVCodecContext), c_void_p, c_int, c_int)),
        ('rtp_payload_size', c_int),  # Deprecated. Removed in 59.
        ('mv_bits', c_int),  # Deprecated. Removed in 59.
        ('header_bits', c_int),  # Deprecated. Removed in 59.
        ('i_tex_bits', c_int),  # Deprecated. Removed in 59.
        ('p_tex_bits', c_int),  # Deprecated. Removed in 59.
        ('i_count', c_int),  # Deprecated. Removed in 59.
        ('p_count', c_int),  # Deprecated. Removed in 59.
        ('skip_count', c_int),  # Deprecated. Removed in 59.
        ('misc_bits', c_int),  # Deprecated. Removed in 59.
        ('frame_bits', c_int),  # Deprecated. Removed in 59.
        ('stats_out', c_char_p),
        ('stats_in', c_char_p),
        ('workaround_bugs', c_int),
        ('strict_std_compliance', c_int),
        ('error_concealment', c_int),
        ('debug', c_int),
        ('err_recognition', c_int),
        ('reordered_opaque', c_int64),
        ('hwaccel', POINTER(AVHWAccel)),
        ('hwaccel_context', c_void_p),
        ('error', c_uint64 * AV_NUM_DATA_POINTERS),
        ('dct_algo', c_int),
        ('idct_algo', c_int),
        ('bits_per_coded_sample', c_int),
        ('bits_per_raw_sample', c_int),
        ('lowres', c_int),
        ('coded_frame', POINTER(AVFrame)),  # Deprecated. Removed in 59.
        ('thread_count', c_int),
        ('thread_type', c_int),
        ('active_thread_type', c_int),
        ('thread_safe_callbacks', c_int),
        ('execute', CFUNCTYPE(c_int,
                              POINTER(AVCodecContext),
                              CFUNCTYPE(c_int, POINTER(AVCodecContext), c_void_p),
                              c_void_p, c_int, c_int, c_int)),
        ('execute2', CFUNCTYPE(c_int,
                               POINTER(AVCodecContext),
                               CFUNCTYPE(c_int, POINTER(AVCodecContext), c_void_p, c_int, c_int),
                               c_void_p, c_int, c_int)),
        ('nsse_weight', c_int),
        ('profile', c_int),
        ('level', c_int),
        ('skip_loop_filter', c_int),
        ('skip_idct', c_int),
        ('skip_frame', c_int),
        ('subtitle_header', POINTER(c_uint8)),
        ('subtitle_header_size', c_int),
        ('vbv_delay', c_uint64),  # Deprecated. Removed in 59.
        ('side_data_only_packets', c_int),  # Deprecated. Removed in 59.
        ('initial_padding', c_int),
        ('framerate', AVRational),
        # !
        ('sw_pix_fmt', c_int),
        ('pkt_timebase', AVRational),
        ('codec_dexcriptor', AVCodecDescriptor),
        ('pts_correction_num_faulty_pts', c_int64),
        ('pts_correction_num_faulty_dts', c_int64),
        ('pts_correction_last_pts', c_int64),
        ('pts_correction_last_dts', c_int64),
        ('sub_charenc', c_char_p),
        ('sub_charenc_mode', c_int),
        ('skip_alpha', c_int),
        ('seek_preroll', c_int),
        ('debug_mv', c_int),
        ('chroma_intra_matrix', POINTER(c_uint16)),
        ('dump_separator', POINTER(c_uint8)),
        ('codec_whitelist', c_char_p),
        ('properties', c_uint),
        ('coded_side_data', POINTER(AVPacketSideData)),
        ('nb_coded_side_data', c_int),
        ('hw_frames_ctx', POINTER(AVBufferRef)),
        ('sub_text_format', c_int),
        ('trailing_padding', c_int),
        ('max_pixels', c_int64),
        ('hw_device_ctx', POINTER(AVBufferRef)),
        ('hwaccel_flags', c_int),
        ('apply_cropping', c_int),
        ('extra_hw_frames', c_int)
    ]

    compat.add_version_changes('avcodec', 58, AVCodecContext, AVCodecContext_Fields, removals=None)

    for compat_ver in (59, 60):
        compat.add_version_changes('avcodec', compat_ver, AVCodecContext, AVCodecContext_Fields,
            removals=('b_frame_strategy', 'mpeg_quant', 'prediction_method', 'pre_me', 'scenechange_threshold',
                          'noise_reduction', 'me_penalty_compensation', 'brd_scale', 'chromaoffset', 'b_sensitivity',
                          'refcounted_frames', 'coder_type', 'context_model', 'coder_type', 'context_model',
                          'frame_skip_threshold', 'frame_skip_factor', 'frame_skip_exp', 'frame_skip_cmp',
                          'min_prediction_order', 'max_prediction_order', 'timecode_frame_start', 'rtp_callback',
                          'rtp_payload_size', 'mv_bits', 'header_bits', 'i_tex_bits', 'p_tex_bits', 'i_count', 'p_count',
                          'skip_count', 'misc_bits', 'frames_bits', 'coded_frame', 'vbv_delay', 'side_data_only_packets')
        )

AV_CODEC_ID_VP8 = 139
AV_CODEC_ID_VP9 = 167

avcodec.av_packet_unref.argtypes = [POINTER(AVPacket)]
avcodec.av_packet_free.argtypes = [POINTER(POINTER(AVPacket))]
avcodec.av_packet_clone.restype = POINTER(AVPacket)
avcodec.av_packet_clone.argtypes = [POINTER(AVPacket)]
avcodec.av_packet_move_ref.argtypes = [POINTER(AVPacket), POINTER(AVPacket)]
avcodec.avcodec_find_decoder.restype = POINTER(AVCodec)
avcodec.avcodec_find_decoder.argtypes = [c_int]
AVDictionary = libavutil.AVDictionary
avcodec.avcodec_open2.restype = c_int
avcodec.avcodec_open2.argtypes = [POINTER(AVCodecContext),
                                  POINTER(AVCodec),
                                  POINTER(POINTER(AVDictionary))]
avcodec.avcodec_free_context.argtypes = [POINTER(POINTER(AVCodecContext))]
avcodec.av_packet_alloc.restype = POINTER(AVPacket)
avcodec.av_init_packet.argtypes = [POINTER(AVPacket)]

avcodec.avcodec_receive_frame.restype = c_int
avcodec.avcodec_receive_frame.argtypes = [POINTER(AVCodecContext), POINTER(AVFrame)]

avcodec.avcodec_send_packet.restype = c_int
avcodec.avcodec_send_packet.argtypes = [POINTER(AVCodecContext), POINTER(AVPacket)]

avcodec.avcodec_flush_buffers.argtypes = [POINTER(AVCodecContext)]
avcodec.avcodec_alloc_context3.restype = POINTER(AVCodecContext)
avcodec.avcodec_alloc_context3.argtypes = [POINTER(AVCodec)]
avcodec.avcodec_free_context.argtypes = [POINTER(POINTER(AVCodecContext))]
avcodec.avcodec_parameters_to_context.restype = c_int
avcodec.avcodec_parameters_to_context.argtypes = [POINTER(AVCodecContext),
                                                  POINTER(AVCodecParameters)]
avcodec.avcodec_get_name.restype = c_char_p
avcodec.avcodec_get_name.argtypes = [c_int]
avcodec.avcodec_find_decoder_by_name.restype = POINTER(AVCodec)
avcodec.avcodec_find_decoder_by_name.argtypes = [c_char_p]

__all__ = [
    'avcodec',
    'FF_INPUT_BUFFER_PADDING_SIZE',
    'AVPacket',
    'AVCodecContext',
    'AV_CODEC_ID_VP8',
    'AV_CODEC_ID_VP9',
]



#!/usr/bin/env python
"""Demonstrates one way of fixing the display resolution to a certain
size, but rendering to the full screen.

The method used in this example is:

1. Create a Framebuffer object, a Texture, and a depth buffer.
2. Attach the Texture and depth buffer to the Framebuffer.
3. Bind the Framebuffer as the current render target.
4. Render the scene using any OpenGL functions (here, just a shape).
5. Unbind the Framebuffer, and blit the Texture scaled to fill the screen.
"""

from math import floor

import pyglet

from pyglet.gl import *


class FixedResolution:
    def __init__(self, window, width, height):
        self.window = window
        self.window.set_minimum_size(width, height)
        self.width = width
        self.height = height

        self._render_area = 0, 0, 0, 0, 0
        self._mouse_scaling = 0, 0, 1.0, 1.0

        self.framebuffer = pyglet.image.Framebuffer()
        self._color_buffer = pyglet.image.Texture.create(width, height, min_filter=GL_NEAREST, mag_filter=GL_NEAREST)
        self._depth_buffer = pyglet.image.buffer.Renderbuffer(width, height, GL_DEPTH_COMPONENT)
        self.framebuffer.attach_texture(self._color_buffer)
        self.framebuffer.attach_renderbuffer(self._depth_buffer, attachment=GL_DEPTH_ATTACHMENT)

        self.window.push_handlers(self)

    def get_mouse_scale(self, x, y, dx, dy):
        offset_x, offset_y, scale_w, scale_h = self._mouse_scaling
        scaled_x = floor((x - offset_x) * scale_w)
        scaled_y = floor((y - offset_y) * scale_h)
        scaled_dx = floor(dx * scale_w)
        scaled_dy = floor(dy * scale_h)
        return scaled_x, scaled_y, scaled_dx, scaled_dy

    def on_resize(self, new_screen_width, new_screen_height):
        aspect_ratio = self.width / self.height
        aspect_width = new_screen_width
        aspect_height = aspect_width / aspect_ratio + 0.5
        if aspect_height > new_screen_height:
            aspect_height = new_screen_height
            aspect_width = aspect_height * aspect_ratio + 0.5

        offset_x = int((new_screen_width / 2) - (aspect_width / 2))
        offset_y = int((new_screen_height / 2) - (aspect_height / 2))
        render_width = int(aspect_width)
        render_height = int(aspect_height)

        self._render_area = offset_x, offset_y, 0, render_width, render_height

        # mouse resolution scaling:
        width_scale = self.width / new_screen_width
        height_scale = self.height / new_screen_height
        # optional scaling for pillar/letterboxing:
        pillar_scale = new_screen_width / render_width
        letter_scale = new_screen_height / render_height

        self._mouse_scaling = offset_x, offset_y, width_scale * pillar_scale, height_scale * letter_scale

    def __enter__(self):
        self.framebuffer.bind()
        self.window.clear()

    def __exit__(self, *unused):
        self.framebuffer.unbind()
        self._color_buffer.blit(*self._render_area)

    def begin(self):
        self.__enter__()

    def end(self):
        self.__exit__()


###################################
# Simple program using the Viewport:
###################################

if __name__ == '__main__':
    window = pyglet.window.Window(960, 540, resizable=True)

    # Create an instance of the FixedResolution class. Use
    # 320x180 resolution to make the effect completely obvious:
    fixed_res = FixedResolution(window, width=320, height=180)


    @window.event
    def on_draw():
        window.clear()

        # The FixedResolution instance can be used as a context manager:
        with fixed_res:
            rectangle.draw()

        # # Alternatively, you can do it manually:
        # fixed_res.begin()
        # rectangle.draw()
        # fixed_res.end()


    # Create a simple Rectangle to show the effect
    rectangle = pyglet.shapes.Rectangle(x=160, y=90, color=(200, 50, 50), width=100, height=100)
    rectangle.anchor_position = 50, 50

    # Combine update & drawing to avoid stutter from mismatched update rates
    def update(dt):
        global rectangle
        rectangle.rotation += dt * 10

        # This method automatically calls any on_draw method registered
        # using @window.event, as we did above.
        window.draw(dt)


    # Create a simple Rectangle to show the effect
    rectangle = pyglet.shapes.Rectangle(x=160, y=90, color=(200, 50, 50), width=100, height=100)
    rectangle.anchor_position = 50, 50


# Call the combined update & redraw function at 60 FPS
pyglet.clock.schedule_interval(update, 1/60)


# Start the example
pyglet.app.run()
